@@ -28,7 +28,6 @@
 #include <math.h>
 #include <gusb.h>
 #include <libsoup/soup.h>
-#include <stdlib.h>
 #include <lcms2.h>
 
 #include "ch-common.h"
@@ -53,6 +52,7 @@ typedef struct {
 	GHashTable	*hash;
 	guint32		 serial_number;
 	gboolean	 needs_repair;
+	gboolean	 force_repair;
 } ChCcmxPrivate;
 
 enum {
@@ -65,9 +65,10 @@ enum {
 
 static void	 ch_ccmx_get_calibration_idx		(ChCcmxPrivate *priv);
 static void	 ch_ccmx_refresh_calibration_data	(ChCcmxPrivate *priv);
-static gboolean	 ch_ccmx_set_calibration_file		(ChCcmxPrivate *priv,
+static gboolean	 ch_ccmx_set_calibration_data		(ChCcmxPrivate *priv,
 							 guint16 cal_idx,
-							 const gchar *filename,
+							 const guint8 *ccmx_data,
+							 gsize ccmx_size,
 							 GError **error);
 
 /**
@@ -233,6 +234,9 @@ ch_ccmx_add_local_file (ChCcmxPrivate *priv,
 	type_tmp = cmsIT8GetProperty (ccmx, "TYPE_LCD");
 	if (g_strcmp0 (type_tmp, "YES") == 0)
 		types += CH_CALIBRATION_TYPE_LCD;
+	type_tmp = cmsIT8GetProperty (ccmx, "TYPE_LED");
+	if (g_strcmp0 (type_tmp, "YES") == 0)
+		types += CH_CALIBRATION_TYPE_LED;
 	type_tmp = cmsIT8GetProperty (ccmx, "TYPE_CRT");
 	if (g_strcmp0 (type_tmp, "YES") == 0)
 		types += CH_CALIBRATION_TYPE_CRT;
@@ -261,6 +265,22 @@ ch_ccmx_add_local_file (ChCcmxPrivate *priv,
 			     g_strdup (description),
 			     GINT_TO_POINTER (1));
 
+	/* is suitable for LED */
+	list_store = GTK_LIST_STORE (gtk_builder_get_object (priv->builder, "liststore_led"));
+	if ((types & CH_CALIBRATION_TYPE_LED) > 0) {
+		ret = ch_ccmx_find_by_desc (GTK_TREE_MODEL (list_store),
+					    &iter,
+					    description);
+		if (!ret)
+			gtk_list_store_append (list_store, &iter);
+		gtk_list_store_set (list_store, &iter,
+				    COLUMN_DESCRIPTION, description,
+				    COLUMN_INDEX, -1,
+				    COLUMN_TYPE, "web-browser",
+				    COLUMN_LOCAL_FILENAME, filename,
+				    -1);
+	}
+
 	/* is suitable for CRT */
 	list_store = GTK_LIST_STORE (gtk_builder_get_object (priv->builder, "liststore_crt"));
 	if ((types & CH_CALIBRATION_TYPE_CRT) > 0) {
@@ -381,7 +401,6 @@ ch_ccmx_got_factory_calibration_cb (SoupSession *session,
 {
 	ChCcmxPrivate *priv = (ChCcmxPrivate *) user_data;
 	gboolean ret;
-	gchar *basename = NULL;
 	gchar *location = NULL;
 	GError *error = NULL;
 	SoupURI *uri;
@@ -406,35 +425,19 @@ ch_ccmx_got_factory_calibration_cb (SoupSession *session,
 		goto out;
 	}
 
-	/* write file */
-	uri = soup_message_get_uri (msg);
-	basename = g_path_get_basename (soup_uri_get_path (uri));
-	location = g_build_filename (g_get_tmp_dir (),
-				     basename,
-				     NULL);
-	ret = g_file_set_contents (location,
-				   msg->response_body->data,
-				   msg->response_body->length,
-				   &error);
-	if (!ret) {
-		ch_ccmx_error_dialog (priv,
-				      _("Failed to write file"),
-				      error->message);
-		g_error_free (error);
-		goto out;
-	}
-
 	/* update UI */
-	ret = ch_ccmx_set_calibration_file (priv, 0, location, &error);
+	ret = ch_ccmx_set_calibration_data (priv, 0,
+					    (const guint8 *) msg->response_body->data,
+					    (gsize) msg->response_body->length,
+					    &error);
 	if (!ret) {
 		ch_ccmx_error_dialog (priv,
-				       _("Failed to load file"),
+				       _("Failed to load data"),
 				       error->message);
 		g_error_free (error);
 		goto out;
 	}
 out:
-	g_free (basename);
 	g_free (location);
 }
 
@@ -488,10 +491,10 @@ out:
 }
 
 /**
- * ch_ccmx_device_repair_cb:
+ * ch_ccmx_device_needs_repair_cb:
  **/
 static void
-ch_ccmx_device_repair_cb (GtkDialog *dialog,
+ch_ccmx_device_needs_repair_cb (GtkDialog *dialog,
 			  GtkResponseType response_id,
 			  ChCcmxPrivate *priv)
 {
@@ -514,10 +517,10 @@ out:
 }
 
 /**
- * ch_ccmx_device_repair:
+ * ch_ccmx_device_needs_repair:
  **/
 static void
-ch_ccmx_device_repair (ChCcmxPrivate *priv)
+ch_ccmx_device_needs_repair (ChCcmxPrivate *priv)
 {
 	const gchar *message;
 	GtkWidget *dialog;
@@ -544,7 +547,43 @@ ch_ccmx_device_repair (ChCcmxPrivate *priv)
 
 	/* wait async */
 	g_signal_connect (dialog, "response",
-			  G_CALLBACK (ch_ccmx_device_repair_cb),
+			  G_CALLBACK (ch_ccmx_device_needs_repair_cb),
+			  priv);
+	gtk_widget_show (dialog);
+}
+
+/**
+ * ch_ccmx_device_force_repair:
+ **/
+static void
+ch_ccmx_device_force_repair (ChCcmxPrivate *priv)
+{
+	const gchar *message;
+	GtkWidget *dialog;
+	GtkWindow *window;
+
+	/* TRANSLATORS: device is broken and needs fixing */
+	message = _("Update the factory calibration values?");
+	window = GTK_WINDOW(gtk_builder_get_object (priv->builder, "dialog_ccmx"));
+	dialog = gtk_message_dialog_new (window,
+					 GTK_DIALOG_MODAL,
+					 GTK_MESSAGE_QUESTION,
+					 GTK_BUTTONS_NONE,
+					 /* TRANSLATORS: the device has an
+					  * updated calibration matrix */
+					 "%s", _("Device calibration update"));
+	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
+						  "%s", message);
+
+	/* TRANSLATORS: this is a button */
+	gtk_dialog_add_button (GTK_DIALOG (dialog), _("Cancel"), GTK_RESPONSE_NO);
+
+	/* TRANSLATORS: this is a button */
+	gtk_dialog_add_button (GTK_DIALOG (dialog), _("Repair"), GTK_RESPONSE_YES);
+
+	/* wait async */
+	g_signal_connect (dialog, "response",
+			  G_CALLBACK (ch_ccmx_device_needs_repair_cb),
 			  priv);
 	gtk_widget_show (dialog);
 }
@@ -588,13 +627,20 @@ ch_ccmx_get_calibration_map_cb (GObject *source,
 	ch_ccmx_set_combo_from_index (GTK_COMBO_BOX (widget), priv->calibration_map[1]);
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "combobox_projector"));
 	ch_ccmx_set_combo_from_index (GTK_COMBO_BOX (widget), priv->calibration_map[2]);
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "combobox_led"));
+	ch_ccmx_set_combo_from_index (GTK_COMBO_BOX (widget), priv->calibration_map[3]);
 
 	/* we've setup */
 	priv->done_get_cal = TRUE;
 
 	/* offer to repair the device */
 	if (priv->needs_repair)
-		ch_ccmx_device_repair (priv);
+		ch_ccmx_device_needs_repair (priv);
+	else if (priv->force_repair) {
+		ch_ccmx_device_force_repair (priv);
+		/* Force repair only once */
+		priv->force_repair = FALSE;
+	}
 out:
 	return;
 }
@@ -644,6 +690,22 @@ ch_ccmx_get_calibration_cb (GObject *source,
 				    -1);
 	}
 
+	/* is suitable for LED */
+	list_store = GTK_LIST_STORE (gtk_builder_get_object (priv->builder, "liststore_led"));
+	if ((types & CH_CALIBRATION_TYPE_LED) > 0) {
+		ret = ch_ccmx_find_by_desc (GTK_TREE_MODEL (list_store),
+					    &iter,
+					    description);
+		if (!ret)
+			gtk_list_store_append (list_store, &iter);
+		gtk_list_store_set (list_store, &iter,
+				    COLUMN_DESCRIPTION, description,
+				    COLUMN_INDEX, priv->ccmx_idx,
+				    COLUMN_TYPE, NULL,
+				    COLUMN_LOCAL_FILENAME, NULL,
+				    -1);
+	}
+
 	/* is suitable for CRT */
 	list_store = GTK_LIST_STORE (gtk_builder_get_object (priv->builder, "liststore_crt"));
 	if ((types & CH_CALIBRATION_TYPE_CRT) > 0) {
@@ -849,36 +911,30 @@ out:
 }
 
 /**
- * ch_ccmx_set_calibration_file:
+ * ch_ccmx_set_calibration_data:
  **/
 static gboolean
-ch_ccmx_set_calibration_file (ChCcmxPrivate *priv,
+ch_ccmx_set_calibration_data (ChCcmxPrivate *priv,
 			      guint16 cal_idx,
-			      const gchar *filename,
+			      const guint8 *ccmx_data,
+			      gsize ccmx_size,
 			      GError **error)
 {
 	cmsHANDLE ccmx = NULL;
 	const gchar *description;
 	const gchar *sheet_type;
 	const gchar *type_tmp;
-	gboolean ret;
-	gchar *ccmx_data = NULL;
-	gsize ccmx_size;
+	gboolean ret = TRUE;
+	gboolean type_factory = FALSE;
 	guint8 buffer[62];
 	guint8 types = 0;
 	guint i, j;
 
-	/* load file */
-	ret = g_file_get_contents (filename,
-				   &ccmx_data,
-				   &ccmx_size,
-				   error);
-	if (!ret)
-		goto out;
-	ccmx = cmsIT8LoadFromMem (NULL, ccmx_data, ccmx_size);
+	/* load from a blob, as lcms sucks at reading files */
+	ccmx = cmsIT8LoadFromMem (NULL, (void *) ccmx_data, ccmx_size);
 	if (ccmx == NULL) {
 		ret = FALSE;
-		g_set_error (error, 1, 0, "Cannot open %s", filename);
+		g_set_error_literal (error, 1, 0, "Cannot open CCMX");
 		goto out;
 	}
 
@@ -886,8 +942,8 @@ ch_ccmx_set_calibration_file (ChCcmxPrivate *priv,
 	sheet_type = cmsIT8GetSheetType (ccmx);
 	if (g_strcmp0 (sheet_type, "CCMX   ") != 0) {
 		ret = FALSE;
-		g_set_error (error, 1, 0, "%s is not a CCMX file [%s]",
-			     filename, sheet_type);
+		g_set_error (error, 1, 0, "file is not a CCMX file [%s]",
+			     sheet_type);
 		goto out;
 	}
 
@@ -901,14 +957,20 @@ ch_ccmx_set_calibration_file (ChCcmxPrivate *priv,
 	}
 
 	/* get the types */
+	type_tmp = cmsIT8GetProperty (ccmx, "TYPE_FACTORY");
+	if (g_strcmp0 (type_tmp, "YES") == 0)
+		type_factory = TRUE;
 	type_tmp = cmsIT8GetProperty (ccmx, "TYPE_LCD");
-	if (g_strcmp0 (type_tmp, "YES") == 0)
+	if (type_factory || g_strcmp0 (type_tmp, "YES") == 0)
 		types += CH_CALIBRATION_TYPE_LCD;
+	type_tmp = cmsIT8GetProperty (ccmx, "TYPE_LED");
+	if (type_factory || g_strcmp0 (type_tmp, "YES") == 0)
+		types += CH_CALIBRATION_TYPE_LED;
 	type_tmp = cmsIT8GetProperty (ccmx, "TYPE_CRT");
-	if (g_strcmp0 (type_tmp, "YES") == 0)
+	if (type_factory || g_strcmp0 (type_tmp, "YES") == 0)
 		types += CH_CALIBRATION_TYPE_CRT;
 	type_tmp = cmsIT8GetProperty (ccmx, "TYPE_PROJECTOR");
-	if (g_strcmp0 (type_tmp, "YES") == 0)
+	if (type_factory || g_strcmp0 (type_tmp, "YES") == 0)
 		types += CH_CALIBRATION_TYPE_PROJECTOR;
 
 	/* write the index */
@@ -946,12 +1008,41 @@ ch_ccmx_set_calibration_file (ChCcmxPrivate *priv,
 	if (!ret)
 		goto out;
 out:
-	g_free (ccmx_data);
 	if (ccmx != NULL)
 		cmsIT8Free (ccmx);
 	return ret;
 }
 
+/**
+ * ch_ccmx_set_calibration_file:
+ **/
+static gboolean
+ch_ccmx_set_calibration_file (ChCcmxPrivate *priv,
+			      guint16 cal_idx,
+			      const gchar *filename,
+			      GError **error)
+{
+	gboolean ret;
+	gchar *ccmx_data = NULL;
+	gsize ccmx_size;
+
+	/* load local file */
+	ret = g_file_get_contents (filename,
+				   &ccmx_data,
+				   &ccmx_size,
+				   error);
+	if (!ret)
+		goto out;
+
+	ret = ch_ccmx_set_calibration_data (priv, 0,
+					    (guint8 *)ccmx_data,
+                        ccmx_size,
+					    error);
+out:
+	g_free (ccmx_data);
+	return ret;
+}
+
 /**
  * ch_ccmx_import_button_cb:
  **/
@@ -1510,6 +1601,13 @@ ch_ccmx_startup_cb (GApplication *application, ChCcmxPrivate *priv)
 	g_signal_connect (widget, "changed",
 			  G_CALLBACK (ch_ccmx_combo_changed_cb), priv);
 	ch_ccmx_set_combo_simple_text (widget);
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "combobox_led"));
+	g_object_set_data (G_OBJECT (widget),
+			   "colorhug-ccmx-idx",
+			   GINT_TO_POINTER (3));
+	g_signal_connect (widget, "changed",
+			  G_CALLBACK (ch_ccmx_combo_changed_cb), priv);
+	ch_ccmx_set_combo_simple_text (widget);
 
 	/* setup USB image */
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "image_usb"));
@@ -1614,6 +1712,7 @@ main (int argc, char **argv)
 	ChCcmxPrivate *priv;
 	gboolean ret;
 	gboolean verbose = FALSE;
+	gboolean force_repair = FALSE;
 	GError *error = NULL;
 	GOptionContext *context;
 	int status = 0;
@@ -1621,6 +1720,9 @@ main (int argc, char **argv)
 		{ "verbose", 'v', 0, G_OPTION_ARG_NONE, &verbose,
 			/* TRANSLATORS: command line option */
 			_("Show extra debugging information"), NULL },
+		{ "repair", 'r', 0, G_OPTION_ARG_NONE, &force_repair,
+			/* TRANSLATORS: command line option */
+			_("Repair the factory calibration matrix"), NULL },
 		{ NULL}
 	};
 
@@ -1648,6 +1750,7 @@ main (int argc, char **argv)
 
 	priv = g_new0 (ChCcmxPrivate, 1);
 	priv->needs_repair = TRUE;
+	priv->force_repair = force_repair;
 	priv->usb_ctx = g_usb_context_new (NULL);
 	priv->hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
 	priv->device_list = g_usb_device_list_new (priv->usb_ctx);
