@@ -29,6 +29,11 @@
 #include "ch-common.h"
 #include "ch-math.h"
 
+/* the default timeout */
+#define CH_DEVICE_USB_TIMEOUT		5000 /* ms */
+#define CH_DEVICE_DETERMINANT_AVE	21.53738
+#define CH_DEVICE_DETERMINANT_ERROR	10.00000
+
 /**
  * ch_strerror:
  **/
@@ -98,6 +103,60 @@ ch_strerror (ChError error_enum)
 	return str;
 }
 
+/**
+ * ch_color_select_to_string:
+ **/
+const gchar *
+ch_color_select_to_string (ChColorSelect color_select)
+{
+	const char *str = NULL;
+	switch (color_select) {
+	case CH_COLOR_SELECT_BLUE:
+		str = "Blue";
+		break;
+	case CH_COLOR_SELECT_RED:
+		str = "Red";
+		break;
+	case CH_COLOR_SELECT_GREEN:
+		str = "Green";
+		break;
+	case CH_COLOR_SELECT_WHITE:
+		str = "White";
+		break;
+	default:
+		str = "Unknown";
+		break;
+	}
+	return str;
+}
+
+/**
+ * ch_multiplier_to_string:
+ **/
+const gchar *
+ch_multiplier_to_string (ChFreqScale multiplier)
+{
+	const char *str = NULL;
+	switch (multiplier) {
+	case CH_FREQ_SCALE_0:
+		str = "0%";
+		break;
+	case CH_FREQ_SCALE_2:
+		str = "2%";
+		break;
+	case CH_FREQ_SCALE_20:
+		str = "20%";
+		break;
+	case CH_FREQ_SCALE_100:
+		str = "100%";
+		break;
+	default:
+		str = "Unknown%";
+		break;
+	}
+	return str;
+}
+
 /**
  * ch_command_to_string:
  **/
@@ -139,12 +198,30 @@ ch_command_to_string (guint8 cmd)
 	case CH_CMD_SET_SERIAL_NUMBER:
 		str = "set-serial-number";
 		break;
+	case CH_CMD_GET_OWNER_NAME:
+		str = "get-owner-name";
+		break;
+	case CH_CMD_SET_OWNER_NAME:
+		str = "set-owner-name";
+		break;
+	case CH_CMD_GET_OWNER_EMAIL:
+		str = "get-owner-name";
+		break;
+	case CH_CMD_SET_OWNER_EMAIL:
+		str = "set-owner-email";
+		break;
 	case CH_CMD_GET_LEDS:
 		str = "get-leds";
 		break;
 	case CH_CMD_SET_LEDS:
 		str = "set-leds";
 		break;
+	case CH_CMD_GET_PCB_ERRATA:
+		str = "get-pcb-errata";
+		break;
+	case CH_CMD_SET_PCB_ERRATA:
+		str = "set-pcb-errata";
+		break;
 	case CH_CMD_GET_DARK_OFFSETS:
 		str = "get-dark-offsets";
 		break;
@@ -221,9 +298,6 @@ ch_print_data_buffer (const gchar *title,
 	g_print ("%c[%dm\n", 0x1B, 0);
 }
 
-/* the default timeout */
-#define CH_DEVICE_USB_TIMEOUT	2000
-
 typedef struct {
 	GUsbDevice		*device;
 	GCancellable		*cancellable;
@@ -319,7 +393,7 @@ ch_device_reply_cb (GObject *source_object,
 		error_enum = helper->buffer[CH_BUFFER_OUTPUT_RETVAL];
 		msg = g_strdup_printf ("Invalid read: retval=0x%02x [%s] "
 				       "cmd=0x%02x (expected 0x%x [%s]) "
-				       "len=%li (expected %li)",
+				       "len=%" G_GSIZE_FORMAT " (expected %" G_GSIZE_FORMAT ")",
 				       error_enum,
 				       ch_strerror (error_enum),
 				       helper->buffer[CH_BUFFER_OUTPUT_CMD],
@@ -860,12 +934,13 @@ out:
 gboolean
 ch_device_cmd_get_calibration (GUsbDevice *device,
 			       guint16 calibration_index,
-			       gdouble *calibration,
+			       CdMat3x3 *calibration,
 			       guint8 *types,
 			       gchar *description,
 			       GError **error)
 {
 	gboolean ret;
+	gdouble *calibration_tmp;
 	guint8 buffer[9*4 + 1 + CH_CALIBRATION_DESCRIPTION_LEN];
 	guint i;
 
@@ -887,9 +962,10 @@ ch_device_cmd_get_calibration (GUsbDevice *device,
 
 	/* convert back into floating point */
 	if (calibration != NULL) {
+		calibration_tmp = cd_mat33_get_data (calibration);
 		for (i = 0; i < 9; i++) {
 			ch_packed_float_to_double ((ChPackedFloat *) &buffer[i*4],
-						   &calibration[i]);
+						   &calibration_tmp[i]);
 		}
 	}
 
@@ -913,12 +989,13 @@ out:
 gboolean
 ch_device_cmd_set_calibration (GUsbDevice *device,
 			       guint16 calibration_index,
-			       const gdouble *calibration,
+			       const CdMat3x3 *calibration,
 			       guint8 types,
 			       const gchar *description,
 			       GError **error)
 {
 	gboolean ret;
+	gdouble *calibration_tmp;
 	guint8 buffer[9*4 + 2 + 1 + CH_CALIBRATION_DESCRIPTION_LEN];
 	guint i;
 
@@ -933,7 +1010,8 @@ ch_device_cmd_set_calibration (GUsbDevice *device,
 
 	/* convert from float to signed value */
 	for (i = 0; i < 9; i++) {
-		ch_double_to_packed_float (calibration[i],
+		calibration_tmp = cd_mat33_get_data (calibration);
+		ch_double_to_packed_float (calibration_tmp[i],
 					   (ChPackedFloat *) &buffer[i*4 + 2]);
 	}
 
@@ -969,15 +1047,20 @@ ch_device_cmd_set_calibration_ccmx (GUsbDevice *device,
 				    const gchar *filename,
 				    GError **error)
 {
+	CdMat3x3 calibration;
 	cmsHANDLE ccmx = NULL;
 	const gchar *description;
 	const gchar *sheet_type;
 	gboolean ret;
+	gdouble *calibration_tmp;
+#if CD_CHECK_VERSION(0,1,17)
+	gdouble det;
+#endif
 	gchar *ccmx_data = NULL;
-	gdouble calibration[9];
 	gsize ccmx_size;
+	guint i;
 
-	g_return_val_if_fail (calibration != NULL, FALSE);
+	g_return_val_if_fail (filename != NULL, FALSE);
 	g_return_val_if_fail (G_USB_IS_DEVICE (device), FALSE);
 	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
@@ -1014,20 +1097,44 @@ ch_device_cmd_set_calibration_ccmx (GUsbDevice *device,
 	}
 
 	/* get the values */
-	calibration[0] = cmsIT8GetDataRowColDbl(ccmx, 0, 0);
-	calibration[1] = cmsIT8GetDataRowColDbl(ccmx, 0, 1);
-	calibration[2] = cmsIT8GetDataRowColDbl(ccmx, 0, 2);
-	calibration[3] = cmsIT8GetDataRowColDbl(ccmx, 1, 0);
-	calibration[4] = cmsIT8GetDataRowColDbl(ccmx, 1, 1);
-	calibration[5] = cmsIT8GetDataRowColDbl(ccmx, 1, 2);
-	calibration[6] = cmsIT8GetDataRowColDbl(ccmx, 2, 0);
-	calibration[7] = cmsIT8GetDataRowColDbl(ccmx, 2, 1);
-	calibration[8] = cmsIT8GetDataRowColDbl(ccmx, 2, 2);
+	calibration.m00 = cmsIT8GetDataRowColDbl(ccmx, 0, 0);
+	calibration.m01 = cmsIT8GetDataRowColDbl(ccmx, 0, 1);
+	calibration.m02 = cmsIT8GetDataRowColDbl(ccmx, 0, 2);
+	calibration.m10 = cmsIT8GetDataRowColDbl(ccmx, 1, 0);
+	calibration.m11 = cmsIT8GetDataRowColDbl(ccmx, 1, 1);
+	calibration.m12 = cmsIT8GetDataRowColDbl(ccmx, 1, 2);
+	calibration.m20 = cmsIT8GetDataRowColDbl(ccmx, 2, 0);
+	calibration.m21 = cmsIT8GetDataRowColDbl(ccmx, 2, 1);
+	calibration.m22 = cmsIT8GetDataRowColDbl(ccmx, 2, 2);
+
+	/* check for sanity */
+	calibration_tmp = cd_mat33_get_data (&calibration);
+	for (i = 0; i < 9; i++) {
+		if (calibration_tmp[i] < -10.0f ||
+		    calibration_tmp[i] > 10.0f) {
+			ret = FALSE;
+			g_set_error (error, 1, 0,
+				     "Matrix value %i out of range %f",
+				     i, calibration_tmp[i]);
+			goto out;
+		}
+	}
+
+#if CD_CHECK_VERSION(0,1,17)
+	/* check the scale is correct */
+	det = cd_mat33_determinant (&calibration);
+	if (ABS (det - CH_DEVICE_DETERMINANT_AVE) > CH_DEVICE_DETERMINANT_ERROR) {
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "Matrix determinant out of range: %f", det);
+		goto out;
+	}
+#endif
 
 	/* set to HW */
 	ret = ch_device_cmd_set_calibration (device,
 					     calibration_index,
-					     calibration,
+					     &calibration,
 					     CH_CALIBRATION_TYPE_ALL,
 					     description,
 					     error);
@@ -1282,7 +1389,7 @@ out:
  **/
 gboolean
 ch_device_cmd_get_leds (GUsbDevice *device,
-			ChStatusLed *leds,
+			guint8 *leds,
 			GError **error)
 {
 	gboolean ret;
@@ -1291,12 +1398,14 @@ ch_device_cmd_get_leds (GUsbDevice *device,
 	g_return_val_if_fail (leds != NULL, FALSE);
 	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
+	*leds = 0;
+
 	/* hit hardware */
 	ret = ch_device_write_command (device,
 				       CH_CMD_GET_LEDS,
 				       NULL,	/* buffer in */
 				       0,	/* size of input buffer */
-				       (guint8 *) leds,
+				       leds,
 				       1,	/* size of output buffer */
 				       NULL,	/* cancellable */
 				       error);
@@ -1311,7 +1420,7 @@ out:
  **/
 gboolean
 ch_device_cmd_set_leds (GUsbDevice *device,
-			ChStatusLed leds,
+			guint8 leds,
 			guint8 repeat,
 			guint8 on_time,
 			guint8 off_time,
@@ -1343,6 +1452,65 @@ out:
 	return ret;
 }
 
+/**
+ * ch_device_cmd_get_pcb_errata:
+ **/
+gboolean
+ch_device_cmd_get_pcb_errata (GUsbDevice *device,
+			      guint16 *pcb_errata,
+			      GError **error)
+{
+	gboolean ret;
+
+	g_return_val_if_fail (G_USB_IS_DEVICE (device), FALSE);
+	g_return_val_if_fail (pcb_errata != NULL, FALSE);
+	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+	*pcb_errata = CH_PCB_ERRATA_NONE;
+
+	/* hit hardware */
+	ret = ch_device_write_command (device,
+				       CH_CMD_GET_PCB_ERRATA,
+				       NULL,	/* buffer in */
+				       0,	/* size of input buffer */
+				       (guint8 *) pcb_errata,
+				       sizeof (guint16),
+				       NULL,	/* cancellable */
+				       error);
+	if (!ret)
+		goto out;
+out:
+	return ret;
+}
+
+/**
+ * ch_device_cmd_set_pcb_errata:
+ **/
+gboolean
+ch_device_cmd_set_pcb_errata (GUsbDevice *device,
+			      guint16 pcb_errata,
+			      GError **error)
+{
+	gboolean ret;
+
+	g_return_val_if_fail (G_USB_IS_DEVICE (device), FALSE);
+	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+	/* hit hardware */
+	ret = ch_device_write_command (device,
+				       CH_CMD_SET_PCB_ERRATA,
+				       (const guint8 *) &pcb_errata,
+				       sizeof (guint16),
+				       NULL,	/* buffer out */
+				       0,	/* size of output buffer */
+				       NULL,	/* cancellable */
+				       error);
+	if (!ret)
+		goto out;
+out:
+	return ret;
+}
+
 /**
  * ch_device_cmd_write_eeprom:
  **/
@@ -1774,3 +1942,158 @@ ch_device_cmd_get_hardware_version (GUsbDevice *device,
 out:
 	return ret;
 }
+
+/**
+ * ch_device_cmd_get_owner_name:
+ **/
+gboolean
+ch_device_cmd_get_owner_name (GUsbDevice *device,
+			      gchar *name,
+			      GError **error)
+{
+	gboolean ret;
+
+	g_return_val_if_fail (G_USB_IS_DEVICE (device), FALSE);
+	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+	g_return_val_if_fail (name != NULL, FALSE);
+
+	/* hit hardware */
+	ret = ch_device_write_command (device,
+				       CH_CMD_GET_OWNER_NAME,
+				       NULL,	/* buffer in */
+				       0,	/* size of input buffer */
+				       (guint8 *) name,
+				       sizeof(gchar) * CH_OWNER_LENGTH_MAX,	/* size of output buffer */
+				       NULL,	/* cancellable */
+				       error);
+	if (!ret)
+		goto out;
+	name[CH_OWNER_LENGTH_MAX-1] = 0;
+out:
+	return ret;
+}
+
+/**
+ * ch_device_cmd_set_owner_name:
+ **/
+gboolean
+ch_device_cmd_set_owner_name (GUsbDevice *device,
+			      const gchar *name,
+			      GError **error)
+{
+	gboolean ret;
+	gchar buf[CH_OWNER_LENGTH_MAX];
+
+	g_return_val_if_fail (G_USB_IS_DEVICE (device), FALSE);
+	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+	g_return_val_if_fail (name != NULL, FALSE);
+
+	memset(buf, 0, CH_OWNER_LENGTH_MAX);
+	g_strlcpy(buf, name, CH_OWNER_LENGTH_MAX);
+
+	/* hit hardware */
+	ret = ch_device_write_command (device,
+				       CH_CMD_SET_OWNER_NAME,
+				       (const guint8 *) buf,		/* buffer in */
+				       sizeof(gchar) * CH_OWNER_LENGTH_MAX,	/* size of input buffer */
+				       NULL,
+				       0,	/* size of output buffer */
+				       NULL,	/* cancellable */
+				       error);
+	if (!ret)
+		goto out;
+out:
+	return ret;
+}
+
+/**
+ * ch_device_cmd_get_owner_email:
+ **/
+gboolean
+ch_device_cmd_get_owner_email (GUsbDevice *device,
+			       gchar *email,
+			       GError **error)
+{
+	gboolean ret;
+
+	g_return_val_if_fail (G_USB_IS_DEVICE (device), FALSE);
+	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+	g_return_val_if_fail (email != NULL, FALSE);
+
+	/* hit hardware */
+	ret = ch_device_write_command (device,
+				       CH_CMD_GET_OWNER_EMAIL,
+				       NULL,	/* buffer in */
+				       0,	/* size of input buffer */
+				       (guint8 *) email,
+				       sizeof(gchar) * CH_OWNER_LENGTH_MAX,	/* size of output buffer */
+				       NULL,	/* cancellable */
+				       error);
+	if (!ret)
+		goto out;
+	email[CH_OWNER_LENGTH_MAX-1] = 0;
+out:
+	return ret;
+}
+
+/**
+ * ch_device_cmd_set_owner_email:
+ **/
+gboolean
+ch_device_cmd_set_owner_email (GUsbDevice *device,
+			       const gchar *email,
+			       GError **error)
+{
+	gboolean ret;
+	gchar buf[CH_OWNER_LENGTH_MAX];
+
+	g_return_val_if_fail (G_USB_IS_DEVICE (device), FALSE);
+	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+	g_return_val_if_fail (email != NULL, FALSE);
+
+	memset(buf, 0, CH_OWNER_LENGTH_MAX);
+	g_strlcpy(buf, email, CH_OWNER_LENGTH_MAX);
+
+	/* hit hardware */
+	ret = ch_device_write_command (device,
+				       CH_CMD_SET_OWNER_EMAIL,
+				       (const guint8 *) buf,		/* buffer in */
+				       sizeof(gchar) * CH_OWNER_LENGTH_MAX,	/* size of input buffer */
+				       NULL,
+				       0,	/* size of output buffer */
+				       NULL,	/* cancellable */
+				       error);
+	if (!ret)
+		goto out;
+out:
+	return ret;
+}
+
+/**
+ * ch_device_cmd_take_reading_array:
+ **/
+gboolean
+ch_device_cmd_take_reading_array (GUsbDevice *device,
+				  guint8 *reading_array,
+				  GError **error)
+{
+	gboolean ret;
+
+	g_return_val_if_fail (G_USB_IS_DEVICE (device), FALSE);
+	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+	g_return_val_if_fail (reading_array != NULL, FALSE);
+
+	/* hit hardware */
+	ret = ch_device_write_command (device,
+				       CH_CMD_TAKE_READING_ARRAY,
+				       NULL,		/* buffer in */
+				       0,	/* size of input buffer */
+				       reading_array,
+				       30,	/* size of output buffer */
+				       NULL,	/* cancellable */
+				       error);
+	if (!ret)
+		goto out;
+out:
+	return ret;
+}
