@@ -36,6 +36,9 @@
 #define	CH_USB_HID_EP_OUT			(CH_USB_HID_EP | 0x00)
 #define	CH_USB_HID_EP_SIZE			64
 
+/* constants for ownership tags */
+#define CH_OWNER_LENGTH_MAX			60
+
 /**
  * CH_CMD_GET_COLOR_SELECT:
  *
@@ -132,7 +135,8 @@
  *  0     Can be used with LCD panels
  *  1     Can be used with CRT monitors
  *  2     Can be used with projectors
- *  3-7   Reserved for future use
+ *  3     Can be used with LED panels
+ *  4-7   Reserved for future use
  *
  * IN:  [1:cmd][2:index]
  * OUT: [1:retval][1:cmd][2*9:matrix_value][1:types][23:description]
@@ -229,6 +233,54 @@
  **/
 #define	CH_CMD_SET_DARK_OFFSETS			0x10
 
+/**
+ * CH_CMD_GET_OWNER_NAME:
+ *
+ * Get User's Name
+ *
+ * IN:  [1:cmd]
+ * OUT: [1:retval][1:cmd][60:owner-name]
+ *
+ * This command is only available in firmware mode.
+ **/
+#define	CH_CMD_GET_OWNER_NAME			0x11
+
+/**
+ * CH_CMD_SET_OWNER_NAME:
+ *
+ * Set User's Name
+ *
+ * IN:  [1:cmd][60:owner-name]
+ * OUT: [1:retval][1:cmd]
+ *
+ * This command is only available in firmware mode.
+ **/
+#define	CH_CMD_SET_OWNER_NAME			0x12
+
+/**
+ * CH_CMD_GET_OWNER_EMAIL:
+ *
+ * Get User's Email Address
+ *
+ * IN:  [1:cmd]
+ * OUT: [1:retval][1:cmd][60:owner-email]
+ *
+ * This command is only available in firmware mode.
+ **/
+#define	CH_CMD_GET_OWNER_EMAIL			0x13
+
+/**
+ * CH_CMD_SET_OWNER_NAME:
+ *
+ * Set User's Email Address
+ *
+ * IN:  [1:cmd][60:owner-email]
+ * OUT: [1:retval][1:cmd]
+ *
+ * This command is only available in firmware mode.
+ **/
+#define	CH_CMD_SET_OWNER_EMAIL			0x14
+
 /**
  * CH_CMD_WRITE_EEPROM:
  *
@@ -432,12 +484,12 @@
  * LCD		= 0
  * CRT		= 1
  * Projector	= 2
- * Custom1	= 3
- * Custom2	= 4
- * Custom3	= 5
+ * LED		= 3
+ * Custom1	= 4
+ * Custom2	= 5
  *
  * In the future CustomX may be renamed to another display technology,
- * e.g. LED or e-ink.
+ * e.g. e-ink.
  *
  * IN:  [1:cmd]
  * OUT: [1:retval][1:cmd][6*2:types]
@@ -475,6 +527,57 @@
  **/
 #define	CH_CMD_GET_HARDWARE_VERSION		0x30
 
+/**
+ * CH_CMD_TAKE_READING_ARRAY:
+ *
+ * Takes 30 raw samples and returns them in an array.
+ *
+ * This command can be used to find the optimum delay between patches
+ * by showing a black sample area, then white, and then using this
+ * command to find out how long the actual hardware delay is.
+ *
+ * It can also be used to find out how stable the device or output is
+ * over a small amount of time, typically ~2.5 seconds for the maximum
+ * integral time.
+ *
+ * IN:  [1:cmd]
+ * OUT: [1:retval][1:cmd][30:reading_array]
+ *
+ * This command is available in firmware mode.
+ **/
+#define	CH_CMD_TAKE_READING_ARRAY		0x31
+
+/**
+ * CH_CMD_SET_PCB_ERRATA:
+ *
+ * Sets the board errata value. Board errata is used to correct swapped
+ * LEDs and any future problems discovered that only affect some batches
+ * of hardware version 1.
+ *
+ * The errata bitmask is as follows:
+ * 0x00		= No errata for this PCB
+ * 0x01		= Leds are swapped
+ * 0x02-0xffff	= Reserved for future use
+ *
+ * IN:  [1:cmd][2:pcb_errata]
+ * OUT: [1:retval][1:cmd]
+ *
+ * This command is available in firmware mode.
+ **/
+#define	CH_CMD_SET_PCB_ERRATA			0x32
+
+/**
+ * CH_CMD_GET_PCB_ERRATA:
+ *
+ * Gets the board errata value.
+ *
+ * IN:  [1:cmd]
+ * OUT: [1:retval][1:cmd][2:pcb_errata]
+ *
+ * This command is available in firmware mode.
+ **/
+#define	CH_CMD_GET_PCB_ERRATA			0x33
+
 /* secret code */
 #define	CH_WRITE_EEPROM_MAGIC			"Un1c0rn2"
 
@@ -497,6 +600,7 @@
 #define	CH_CALIBRATION_TYPE_LCD			0x01
 #define	CH_CALIBRATION_TYPE_CRT			0x02
 #define	CH_CALIBRATION_TYPE_PROJECTOR		0x04
+#define	CH_CALIBRATION_TYPE_LED			0x08
 #define	CH_CALIBRATION_TYPE_ALL			0xff
 
 /* approximate sample times */
@@ -510,11 +614,13 @@
 #define	CH_FLASH_ERASE_BLOCK_SIZE		0x400	/* 1024 */
 #define	CH_FLASH_WRITE_BLOCK_SIZE		0x040	/* 64 */
 #define	CH_FLASH_TRANSFER_BLOCK_SIZE		0x020	/* 32 */
+#define	CH_FLASH_RECONNECT_TIMEOUT		2500	/* ms */
 
 /* calibration remapping contants */
 #define	CH_CALIBRATION_INDEX_LCD		(CH_CALIBRATION_MAX + 0)
 #define	CH_CALIBRATION_INDEX_CRT		(CH_CALIBRATION_MAX + 1)
 #define	CH_CALIBRATION_INDEX_PROJECTOR		(CH_CALIBRATION_MAX + 2)
+#define	CH_CALIBRATION_INDEX_LED		(CH_CALIBRATION_MAX + 3)
 
 /* which color to select */
 typedef enum {
@@ -524,7 +630,7 @@ typedef enum {
 	CH_COLOR_SELECT_GREEN
 } ChColorSelect;
 
-/* which color to select */
+/* Led colors: possible bitfield values */
 typedef enum {
 	CH_STATUS_LED_GREEN	= 1 << 0,
 	CH_STATUS_LED_RED	= 1 << 1
@@ -561,9 +667,18 @@ typedef enum {
 	CH_ERROR_LAST
 } ChError;
 
+/* any problems with the PCB */
+typedef enum {
+	CH_PCB_ERRATA_NONE		= 0,
+	CH_PCB_ERRATA_SWAPPED_LEDS	= 1 << 0,
+	CH_PCB_ERRATA_LAST		= 1 << 1
+} ChPcbErrata;
+
 /* prototypes */
 const gchar	*ch_strerror			(ChError	 error_enum);
 const gchar	*ch_command_to_string		(guint8		 cmd);
+const gchar	*ch_multiplier_to_string	(ChFreqScale	 multiplier);
+const gchar	*ch_color_select_to_string	(ChColorSelect	 color_select);
 
 void		ch_device_write_command_async	(GUsbDevice	*device,
 						 guint8		 cmd,
@@ -621,13 +736,13 @@ gboolean	 ch_device_cmd_get_firmware_ver	(GUsbDevice	*device,
 						 GError		**error);
 gboolean	 ch_device_cmd_get_calibration	(GUsbDevice	*device,
 						 guint16	 calibration_index,
-						 gdouble	*calibration,
+						 CdMat3x3	*calibration,
 						 guint8		*types,
 						 gchar		*description,
 						 GError		**error);
 gboolean	 ch_device_cmd_set_calibration	(GUsbDevice	*device,
 						 guint16	 calibration_index,
-						 const gdouble	*calibration,
+						 const CdMat3x3	*calibration,
 						 guint8		 types,
 						 const gchar	*description,
 						 GError		**error);
@@ -657,14 +772,20 @@ gboolean	 ch_device_cmd_set_serial_number (GUsbDevice	*device,
 						 guint32	 serial_number,
 						 GError		**error);
 gboolean	 ch_device_cmd_get_leds		(GUsbDevice	*device,
-						 ChStatusLed	*leds,
+						 guint8		*leds,
 						 GError		**error);
 gboolean	 ch_device_cmd_set_leds		(GUsbDevice	*device,
-						 ChStatusLed	 leds,
+						 guint8		 leds,
 						 guint8		 repeat,
 						 guint8		 on_time,
 						 guint8		 off_time,
 						 GError		**error);
+gboolean	 ch_device_cmd_get_pcb_errata	(GUsbDevice	*device,
+						 guint16	*pcb_errata,
+						 GError		**error);
+gboolean	 ch_device_cmd_set_pcb_errata	(GUsbDevice	*device,
+						 guint16	 pcb_errata,
+						 GError		**error);
 gboolean	 ch_device_cmd_get_dark_offsets	(GUsbDevice	*device,
 						 CdColorRGB	*value,
 						 GError		**error);
@@ -705,5 +826,20 @@ gboolean	 ch_device_cmd_erase_flash	(GUsbDevice	*device,
 						guint16		 address,
 						gsize		 len,
 						GError		**error);
+gboolean	 ch_device_cmd_set_owner_name	(GUsbDevice	*device,
+						 const gchar	*name,
+						 GError	 	**error);
+gboolean	 ch_device_cmd_get_owner_name	(GUsbDevice	*device,
+						 gchar		*name,
+						 GError	 	**error);
+gboolean	 ch_device_cmd_set_owner_email	(GUsbDevice	*device,
+						 const gchar	*email,
+						 GError	 	**error);
+gboolean	 ch_device_cmd_get_owner_email	(GUsbDevice	*device,
+						 gchar		*email,
+						 GError	 	**error);
+gboolean	 ch_device_cmd_take_reading_array (GUsbDevice	*device,
+						 guint8		*reading_array,
+						 GError	 	**error);
 
 #endif
