@@ -28,17 +28,19 @@
 #include <math.h>
 #include <gusb.h>
 #include <libsoup/soup.h>
-#include <stdlib.h>
 
 #include "ch-common.h"
 #include "ch-markdown.h"
+#include "ch-flash-md.h"
 
 /* don't change this unless you want to provide firmware updates */
 #define COLORHUG_FIRMWARE_LOCATION	"http://www.hughski.com/downloads/colorhug/firmware/"
 
 typedef struct {
 	gchar		*filename;
+	gchar		*checksum;
 	GString		*update_details;
+	GString		*warning_details;
 	GtkApplication	*application;
 	GtkBuilder	*builder;
 	guint16		 firmware_version[3];
@@ -67,6 +69,7 @@ static void
 ch_flash_error_dialog (ChFlashPrivate *priv, const gchar *title, const gchar *message)
 {
 	GtkWindow *window;
+	GtkWidget *widget;
 	GtkWidget *dialog;
 
 	window = GTK_WINDOW(gtk_builder_get_object (priv->builder, "dialog_flash"));
@@ -75,10 +78,16 @@ ch_flash_error_dialog (ChFlashPrivate *priv, const gchar *title, const gchar *me
 					 GTK_MESSAGE_ERROR,
 					 GTK_BUTTONS_CLOSE,
 					 "%s", title);
-	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
-						  "%s", message);
+	if (message != NULL) {
+		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
+							  "%s", message);
+	}
 	gtk_dialog_run (GTK_DIALOG (dialog));
 	gtk_widget_destroy (dialog);
+
+	/* close main window */
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "dialog_flash"));
+	gtk_widget_destroy (widget);
 }
 
 /**
@@ -98,12 +107,8 @@ ch_flash_error_do_not_panic (ChFlashPrivate *priv)
 	title = _("Flashing the device failed but the ColorHug is not damaged.");
 	g_string_append_printf (msg, "<b>%s</b>\n", title);
 
-	/* TRANSLATORS: this explains the flashy light thing */
-	title = _("If there were any problems flashing the device, it will enter a bootloader mode.");
-	g_string_append_printf (msg, "%s\n", title);
-
 	/* TRANSLATORS: hopefully we can "fix" this automatically */
-	title = _("Just remove the ColorHug device from the computer, reinsert it and re-run this program.");
+	title = _("Remove the ColorHug device from the computer, reinsert it and re-run this program.");
 	g_string_append_printf (msg, "%s", title);
 
 	/* update UI */
@@ -471,7 +476,8 @@ ch_flash_boot_flash_cb (GObject *source,
 	}
 
 	/* wait one second */
-	g_timeout_add (1000, ch_flash_boot_flash_delay_cb, priv);
+	g_timeout_add (CH_FLASH_RECONNECT_TIMEOUT,
+		       ch_flash_boot_flash_delay_cb, priv);
 out:
 	return;
 }
@@ -497,7 +503,7 @@ ch_flash_read_firmware_chunk (ChFlashPrivate *priv)
 	if (priv->flash_idx < priv->firmware_len) {
 		if (priv->flash_idx + priv->flash_chunk_len > priv->firmware_len)
 			priv->flash_chunk_len = priv->firmware_len - priv->flash_idx;
-		g_debug ("Reading at %04x size %li",
+		g_debug ("Reading at %04x size %" G_GSIZE_FORMAT,
 			 CH_EEPROM_ADDR_RUNCODE + priv->flash_idx,
 			 priv->flash_chunk_len);
 		ch_flash_read_flash (priv,
@@ -562,7 +568,7 @@ ch_flash_write_firmware_chunk (ChFlashPrivate *priv)
 	if (priv->flash_idx < priv->firmware_len) {
 		if (priv->flash_idx + priv->flash_chunk_len > priv->firmware_len)
 			priv->flash_chunk_len = priv->firmware_len - priv->flash_idx;
-		g_debug ("Writing at %04x size %li",
+		g_debug ("Writing at %04x size %" G_GSIZE_FORMAT,
 			 CH_EEPROM_ADDR_RUNCODE + priv->flash_idx,
 			 priv->flash_chunk_len);
 		ch_flash_write_flash (priv,
@@ -573,19 +579,6 @@ ch_flash_write_firmware_chunk (ChFlashPrivate *priv)
 		goto out;
 	};
 
-	/* flush to 64 byte chunk */
-	if ((priv->flash_idx & CH_FLASH_TRANSFER_BLOCK_SIZE) == 0) {
-		priv->flash_idx -= priv->flash_chunk_len;
-		priv->flash_idx += CH_FLASH_TRANSFER_BLOCK_SIZE;
-		g_debug ("Flushing at %04x",
-			 CH_EEPROM_ADDR_RUNCODE + priv->flash_idx);
-		ch_flash_write_flash (priv,
-				      CH_EEPROM_ADDR_RUNCODE + priv->flash_idx,
-				      (guint8 *) priv->firmware_data,
-				      0);
-		goto out;
-	}
-
 	/* update UI */
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "label_status"));
 	/* TRANSLATORS: now we've written the firmware, we have to
@@ -770,20 +763,27 @@ ch_flash_reset_cb (GObject *source,
 	}
 
 	/* wait one second */
-	g_timeout_add (1000, ch_flash_reset_delay_cb, priv);
+	g_timeout_add (CH_FLASH_RECONNECT_TIMEOUT,
+		       ch_flash_reset_delay_cb, priv);
 out:
 	return;
 }
 
 /**
- * ch_flash_reset_into_bootloader:
+ * ch_flash_got_firmware_data:
  **/
 static void
-ch_flash_reset_into_bootloader (ChFlashPrivate *priv)
+ch_flash_got_firmware_data (ChFlashPrivate *priv)
 {
 	const gchar *title;
 	GtkWidget *widget;
 
+	/* we can shortcut as we're already in bootloader mode */
+	if (priv->firmware_version[0] == 0) {
+		ch_flash_set_flash_success_0 (priv);
+		return;
+	}
+
 	/* update UI */
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "label_status"));
 	/* TRANSLATORS: switch from firmware mode into bootloader mode */
@@ -813,8 +813,10 @@ ch_flash_got_firmware_cb (SoupSession *session,
 			  SoupMessage *msg,
 			  gpointer user_data)
 {
-	const gchar *title;
 	ChFlashPrivate *priv = (ChFlashPrivate *) user_data;
+	const gchar *title;
+	gchar *checksum_tmp = NULL;
+	gchar *message = NULL;
 
 	/* we failed */
 	if (!SOUP_STATUS_IS_SUCCESSFUL (msg->status_code)) {
@@ -834,23 +836,30 @@ ch_flash_got_firmware_cb (SoupSession *session,
 		goto out;
 	}
 
+	/* check checksum */
+	checksum_tmp = g_compute_checksum_for_data (G_CHECKSUM_SHA1,
+						    (const guchar *) msg->response_body->data,
+						    msg->response_body->length);
+	if (priv->checksum != NULL &&
+	    g_strcmp0 (priv->checksum, checksum_tmp) != 0) {
+		/* TRANSLATORS: the server gave us an invalid file */
+		title = _("Firmware has incorrect checksum");
+		message = g_strdup_printf ("Expected %s, got %s",
+					   priv->checksum, checksum_tmp);
+		ch_flash_error_dialog (priv, title, message);
+		goto out;
+	}
+
 	/* success */
 	priv->firmware_data = g_new0 (guint8, msg->response_body->length);
 	priv->firmware_len = msg->response_body->length;
 	memcpy (priv->firmware_data,
 		msg->response_body->data,
 		priv->firmware_len);
-
-	/* we can shortcut as we're already in bootloader mode */
-	if (priv->firmware_version[0] == 0) {
-		ch_flash_set_flash_success_0 (priv);
-		goto out;
-	}
-
-	/* reset into the bootloader where we can load the firmware */
-	ch_flash_reset_into_bootloader (priv);
+	ch_flash_got_firmware_data (priv);
 out:
-	return;
+	g_free (message);
+	g_free (checksum_tmp);
 }
 
 /**
@@ -900,6 +909,47 @@ out:
 	return;
 }
 
+/**
+ * ch_flash_show_warning_dialog:
+ **/
+static gboolean
+ch_flash_show_warning_dialog (ChFlashPrivate *priv)
+{
+	const gchar *title;
+	GtkWindow *window;
+	GtkWidget *dialog;
+	gchar *format;
+	GtkResponseType response;
+
+	/* anything to show? */
+	if (priv->warning_details->len == 0)
+		return TRUE;
+
+	/* the update text is markdown formatted */
+	format = ch_markdown_parse (priv->markdown,
+				    priv->warning_details->str);
+	window = GTK_WINDOW(gtk_builder_get_object (priv->builder, "dialog_flash"));
+	/* TRANSLATORS: the long details about all the updates that
+	 * are newer than the version the user has installed */
+	title = _("Warnings about this update");
+	dialog = gtk_message_dialog_new (window,
+					 GTK_DIALOG_MODAL,
+					 GTK_MESSAGE_WARNING,
+					 GTK_BUTTONS_NONE, "%s",
+					 title);
+	/* TRANSLATORS: this is the button text to continue the flash
+	 * after showing the user a warning */
+	gtk_dialog_add_button (GTK_DIALOG (dialog), _("Flash anyway"), GTK_RESPONSE_OK);
+	/* TRANSLATORS: this is the button text to abort the flash process */
+	gtk_dialog_add_button (GTK_DIALOG (dialog), _("Do not flash"), GTK_RESPONSE_CANCEL);
+	gtk_message_dialog_format_secondary_markup (GTK_MESSAGE_DIALOG (dialog), "%s",
+						    format);
+	response = gtk_dialog_run (GTK_DIALOG (dialog));
+	gtk_widget_destroy (dialog);
+	g_free (format);
+	return (response == GTK_RESPONSE_OK);
+}
+
 /**
  * ch_flash_flash_button_cb:
  **/
@@ -909,8 +959,14 @@ ch_flash_flash_button_cb (GtkWidget *widget, ChFlashPrivate *priv)
 	const gchar *title;
 	SoupURI *base_uri = NULL;
 	SoupMessage *msg = NULL;
+	gboolean ret;
 	gchar *uri = NULL;
 
+	/* show the user any warning dialog */
+	ret = ch_flash_show_warning_dialog (priv);
+	if (!ret)
+		goto out;
+
 	/* update UI */
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "button_flash"));
 	gtk_widget_hide (widget);
@@ -991,7 +1047,7 @@ ch_flash_version_is_newer (ChFlashPrivate *priv, const gchar *version)
 	if (g_strv_length (split) != 3)
 		goto out;
 	for (i = 0; i < 3; i++)
-		tmp[i] = atoi (split[i]);
+		tmp[i] = g_ascii_strtoull (split[i], NULL, 10);
 
 	/* check versions */
 	if (ch_flash_get_packed_version (tmp) >
@@ -1045,17 +1101,18 @@ ch_flash_has_updates (ChFlashPrivate *priv)
 }
 
 /**
- * ch_flash_got_manifest_cb:
+ * ch_flash_got_metadata_cb:
  **/
 static void
-ch_flash_got_manifest_cb (SoupSession *session,
+ch_flash_got_metadata_cb (SoupSession *session,
 			  SoupMessage *msg,
 			  gpointer user_data)
 {
-	const gchar *title;
 	ChFlashPrivate *priv = (ChFlashPrivate *) user_data;
-	gchar **lines = NULL;
-	gchar **tmp;
+	ChFlashUpdate *update;
+	const gchar *title;
+	GError *error = NULL;
+	GPtrArray *updates = NULL;
 	guint i;
 
 	/* we failed */
@@ -1076,45 +1133,57 @@ ch_flash_got_manifest_cb (SoupSession *session,
 		goto out;
 	}
 
-	/* write file */
-	lines = g_strsplit (msg->response_body->data, "\n", -1);
-	for (i = 0; lines[i] != NULL; i++) {
-		if (lines[i][0] == '\0')
-			continue;
-		if (lines[i][0] == '\t') {
-			g_string_append_printf (priv->update_details,
-						"%s\n", lines[i] + 1);
-			continue;
-		}
+	/* parse file */
+	updates = ch_flash_md_parse_data (msg->response_body->data,
+					  &error);
+	if (updates == NULL) {
+		/* TRANSLATORS: the XML file was corrupt */
+		title = _("Failed to parse the update metadata");
+		ch_flash_error_dialog (priv, title, error->message);
+		g_error_free (error);
+		goto out;
+	}
+	for (i = 0; i < updates->len; i++) {
+		update = g_ptr_array_index (updates, i);
 
-		/* split up version, date, filename */
-		tmp = g_strsplit (lines[i], "\t", -1);
-		if (g_strv_length (tmp) != 3)
-			continue;
-		if (!ch_flash_version_is_newer (priv, tmp[0])) {
-			g_strfreev (tmp);
+		/* this version is older than what we have now */
+		if (!ch_flash_version_is_newer (priv, update->version))
 			break;
+
+		/* add info text */
+		g_string_append_printf (priv->update_details,
+					"%s\n", update->info->str);
+		g_string_append_printf (priv->warning_details,
+					"%s\n", update->warning->str);
+
+		/* save newest available firmware */
+		if (priv->filename == NULL) {
+			priv->filename = g_strdup (update->filename);
+			priv->checksum = g_strdup (update->checksum);
 		}
-		if (i == 0)
-			priv->filename = g_strdup (tmp[2]);
-		g_strfreev (tmp);
 	}
 
 	/* no updates */
-	if (priv->update_details->len == 0) {
+	if (priv->filename == NULL) {
 		ch_flash_no_updates (priv);
 		goto out;
 	}
 
 	/* remove trailing space */
-	g_string_set_size (priv->update_details,
-			   priv->update_details->len - 1);
+	if (priv->update_details->len > 1) {
+		g_string_set_size (priv->update_details,
+				   priv->update_details->len - 1);
+	}
+	if (priv->warning_details->len > 1) {
+		g_string_set_size (priv->warning_details,
+				   priv->warning_details->len - 1);
+	}
 
 	/* setup UI */
 	ch_flash_has_updates (priv);
 out:
-	g_strfreev (lines);
-	return;
+	if (updates != NULL)
+		g_ptr_array_unref (updates);
 }
 
 /**
@@ -1211,10 +1280,13 @@ ch_flash_got_blacklist_cb (SoupSession *session,
 			   SoupMessage *msg,
 			   gpointer user_data)
 {
-	const gchar *title;
 	ChFlashPrivate *priv = (ChFlashPrivate *) user_data;
+	const gchar *title;
+	gboolean ret;
 	gchar **lines = NULL;
 	gchar *uri = NULL;
+	GError *error = NULL;
+	GtkWidget *widget;
 	guint i;
 	guint tmp;
 	SoupURI *base_uri = NULL;
@@ -1238,16 +1310,38 @@ ch_flash_got_blacklist_cb (SoupSession *session,
 		if (lines[i][0] == '\0')
 			continue;
 		/* check if matches serial number */
-		tmp = atoi (lines[i]);
+		tmp = g_ascii_strtoull (lines[i], NULL, 10);
 		if (tmp == priv->serial_number) {
 			ch_flash_set_lost_or_stolen (priv);
 			goto out;
 		}
 	}
 
+	/* we've manually specified a local firmware file */
+	if (priv->filename != NULL) {
+		/* TRANSLATORS: we've specified a local file */
+		title = _("Flashing firmware file...");
+		widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "label_msg"));
+		gtk_label_set_label (GTK_LABEL (widget), title);
+		ret = g_file_get_contents (priv->filename,
+					   (gchar **) &priv->firmware_data,
+					   &priv->firmware_len,
+					   &error);
+		if (!ret) {
+			/* TRANSLATORS: file read error when reading
+			 * local firmware file */
+			title = _("Failed to load file");
+			ch_flash_error_dialog (priv, title, error->message);
+			g_error_free (error);
+			goto out;
+		}
+		ch_flash_got_firmware_data (priv);
+		goto out;
+	}
+
 	/* get the latest manifest file */
 	uri = g_build_filename (COLORHUG_FIRMWARE_LOCATION,
-				"MANIFEST",
+				"metadata.xml",
 				NULL);
 	base_uri = soup_uri_new (uri);
 
@@ -1262,7 +1356,7 @@ ch_flash_got_blacklist_cb (SoupSession *session,
 
 	/* send sync */
 	soup_session_queue_message (priv->session, msg,
-				    ch_flash_got_manifest_cb, priv);
+				    ch_flash_got_metadata_cb, priv);
 out:
 	if (base_uri != NULL)
 		soup_uri_free (base_uri);
@@ -1608,8 +1702,8 @@ ch_flash_activate_link_cb (GtkLabel *label,
 					 GTK_MESSAGE_INFO,
 					 GTK_BUTTONS_CLOSE, "%s",
 					 title);
-	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog), "%s",
-						  format);
+	gtk_message_dialog_format_secondary_markup (GTK_MESSAGE_DIALOG (dialog), "%s",
+						    format);
 	gtk_dialog_run (GTK_DIALOG (dialog));
 	gtk_widget_destroy (dialog);
 	g_free (format);
@@ -1800,6 +1894,7 @@ main (int argc, char **argv)
 	ChFlashPrivate *priv;
 	gboolean ret;
 	gboolean verbose = FALSE;
+	gchar *filename = NULL;
 	GError *error = NULL;
 	GOptionContext *context;
 	int status = 0;
@@ -1807,6 +1902,9 @@ main (int argc, char **argv)
 		{ "verbose", 'v', 0, G_OPTION_ARG_NONE, &verbose,
 			/* TRANSLATORS: command line option */
 			_("Show extra debugging information"), NULL },
+		{ "filename", 'f', 0, G_OPTION_ARG_STRING, &filename,
+			/* TRANSLATORS: command line option */
+			_("Show extra debugging information"), NULL },
 		{ NULL}
 	};
 
@@ -1832,7 +1930,9 @@ main (int argc, char **argv)
 	g_option_context_free (context);
 
 	priv = g_new0 (ChFlashPrivate, 1);
+	priv->filename = filename;
 	priv->update_details = g_string_new ("");
+	priv->warning_details = g_string_new ("");
 	priv->markdown = ch_markdown_new ();
 	priv->usb_ctx = g_usb_context_new (NULL);
 	priv->device_list = g_usb_device_list_new (priv->usb_ctx);
@@ -1862,6 +1962,8 @@ main (int argc, char **argv)
 	g_object_unref (priv->application);
 	if (priv->update_details != NULL)
 		g_string_free (priv->update_details, TRUE);
+	if (priv->warning_details != NULL)
+		g_string_free (priv->warning_details, TRUE);
 	if (priv->device_list != NULL)
 		g_object_unref (priv->device_list);
 	if (priv->usb_ctx != NULL)
@@ -1873,6 +1975,7 @@ main (int argc, char **argv)
 	if (priv->markdown != NULL)
 		g_object_unref (priv->markdown);
 	g_free (priv->filename);
+	g_free (priv->checksum);
 	g_free (priv->firmware_data);
 	g_free (priv);
 	return status;
