@@ -24,7 +24,6 @@
 #include <glib/gi18n.h>
 #include <gio/gio.h>
 #include <locale.h>
-#include <stdlib.h>
 #include <lcms2.h>
 #include <math.h>
 #include <sqlite3.h>
@@ -177,6 +176,33 @@ out:
 	return ret;
 }
 
+/**
+ * ch_util_get_prompt:
+ **/
+static gboolean
+ch_util_get_prompt (const gchar *question, gboolean defaultyes)
+{
+	gboolean ret = FALSE;
+	gboolean valid = FALSE;
+	gchar value;
+
+	g_print ("%s %s ",
+		 question,
+		 defaultyes ? "[Y/n]" : "[N/y]");
+	while (!valid) {
+		value = getchar ();
+		if (value == 'y' || value == 'Y') {
+			valid = TRUE;
+			ret = TRUE;
+		}
+		if (value == 'n' || value == 'N') {
+			valid = TRUE;
+			ret = FALSE;
+		}
+	}
+	return ret;
+}
+
 /**
  * ch_util_get_color_select:
  **/
@@ -184,7 +210,7 @@ static gboolean
 ch_util_get_color_select (ChUtilPrivate *priv, gchar **values, GError **error)
 {
 	gboolean ret;
-	ChColorSelect color_select;
+	ChColorSelect color_select = 0;
 
 	/* get from HW */
 	ret = ch_device_cmd_get_color_select (priv->device, &color_select, error);
@@ -193,16 +219,10 @@ ch_util_get_color_select (ChUtilPrivate *priv, gchar **values, GError **error)
 
 	switch (color_select) {
 	case CH_COLOR_SELECT_BLUE:
-		g_print ("Blue\n");
-		break;
 	case CH_COLOR_SELECT_RED:
-		g_print ("Red\n");
-		break;
 	case CH_COLOR_SELECT_GREEN:
-		g_print ("Green\n");
-		break;
 	case CH_COLOR_SELECT_WHITE:
-		g_print ("White\n");
+		g_print ("%s\n", ch_color_select_to_string (color_select));
 		break;
 	default:
 		ret = FALSE;
@@ -241,6 +261,79 @@ out:
 	return ret;
 }
 
+/**
+ * ch_util_take_reading_array:
+ **/
+static gboolean
+ch_util_take_reading_array (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gdouble ave = 0.0f;
+	gint i, j;
+	guint8 max = 0;
+	guint8 reading_array[30];
+	gdouble std_dev = 0.0f;
+
+	/* setup HW */
+	ret = ch_device_cmd_set_integral_time (priv->device,
+					       CH_INTEGRAL_TIME_VALUE_MAX,
+					       error);
+	if (!ret)
+		goto out;
+	ret = ch_device_cmd_set_multiplier (priv->device,
+					    CH_FREQ_SCALE_100,
+					    error);
+	if (!ret)
+		goto out;
+	ret = ch_device_cmd_set_color_select (priv->device,
+					      CH_COLOR_SELECT_WHITE,
+					      error);
+	if (!ret)
+		goto out;
+
+	/* get from HW */
+	ret = ch_device_cmd_take_reading_array (priv->device,
+					        reading_array,
+					        error);
+	if (!ret)
+		goto out;
+
+	/* show as a bar graph */
+	for (i = 0; i < 30; i++) {
+		if (reading_array[i] > max)
+			max = reading_array[i];
+		ave += reading_array[i];
+	}
+	ave /= 30;
+	for (i = 0; i < 30; i++) {
+		g_print ("%i.\t%u\t[",
+			 i + 1,
+			 reading_array[i]);
+		for (j = 0; j < reading_array[i]; j++) {
+			if (j == floor (ave)) {
+				g_print ("#");
+				continue;
+			}
+			g_print ("*");
+		}
+		for (j = reading_array[i]; j < max; j++) {
+			if (j == floor (ave)) {
+				g_print (".");
+				continue;
+			}
+			g_print (" ");
+		}
+		g_print ("]\n");
+	}
+
+	/* print standard deviation */
+	for (i = 0; i < 30; i++)
+		std_dev += pow (reading_array[i] - ave, 2);
+	g_print ("Standard deviation: %.03lf\n", sqrt (std_dev / 60));
+out:
+	return ret;
+}
+
 /**
  * ch_util_set_color_select:
  **/
@@ -300,13 +393,9 @@ ch_util_get_multiplier (ChUtilPrivate *priv, gchar **values, GError **error)
 		g_print ("0%% (disabled)\n");
 		break;
 	case CH_FREQ_SCALE_2:
-		g_print ("2%%\n");
-		break;
 	case CH_FREQ_SCALE_20:
-		g_print ("20%%\n");
-		break;
 	case CH_FREQ_SCALE_100:
-		g_print ("100%%\n");
+		g_print ("%s\n", ch_multiplier_to_string (multiplier));
 		break;
 	default:
 		ret = FALSE;
@@ -392,7 +481,7 @@ ch_util_set_integral_time (ChUtilPrivate *priv, gchar **values, GError **error)
 				     "invalid input, expect 'value'");
 		goto out;
 	}
-	integral_time = atoi (values[0]);
+	integral_time = g_ascii_strtoull (values[0], NULL, 10);
 
 	/* set to HW */
 	ret = ch_device_cmd_set_integral_time (priv->device, integral_time, error);
@@ -446,7 +535,7 @@ ch_util_set_calibration_map (ChUtilPrivate *priv,
 		goto out;
 	}
 	for (i = 0; i < 6; i++)
-		calibration_map[i] = atoi (values[i]);
+		calibration_map[i] = g_ascii_strtoull (values[i], NULL, 10);
 
 	/* set to HW */
 	ret = ch_device_cmd_set_calibration_map (priv->device,
@@ -484,13 +573,15 @@ out:
  * ch_util_show_calibration:
  **/
 static void
-ch_util_show_calibration (const gdouble *calibration)
+ch_util_show_calibration (const CdMat3x3 *calibration)
 {
+	gdouble *calibration_tmp;
 	guint i, j;
+	calibration_tmp = cd_mat33_get_data (calibration);
 	for (j = 0; j < 3; j++) {
 		g_print ("( ");
 		for (i = 0; i < 3; i++) {
-			g_print ("%.2f\t", calibration[j*3 + i]);
+			g_print ("%.2f\t", calibration_tmp[j*3 + i]);
 		}
 		g_print (")\n");
 	}
@@ -503,7 +594,7 @@ static gboolean
 ch_util_get_calibration (ChUtilPrivate *priv, gchar **values, GError **error)
 {
 	gboolean ret;
-	gdouble calibration[9];
+	CdMat3x3 calibration;
 	guint16 calibration_index = 0;
 	gchar description[24];
 	guint8 types;
@@ -515,12 +606,12 @@ ch_util_get_calibration (ChUtilPrivate *priv, gchar **values, GError **error)
 				     "invalid input, expect 'calibration_index'");
 		goto out;
 	}
-	calibration_index = atoi (values[0]);
+	calibration_index = g_ascii_strtoull (values[0], NULL, 10);
 
 	/* get from HW */
 	ret = ch_device_cmd_get_calibration (priv->device,
 					     calibration_index,
-					     calibration,
+					     &calibration,
 					     &types,
 					     description,
 					     error);
@@ -531,7 +622,7 @@ ch_util_get_calibration (ChUtilPrivate *priv, gchar **values, GError **error)
 	g_print ("supports CRT: %i\n", (types & 0x02) > 0);
 	g_print ("supports projector: %i\n", (types & 0x04) > 0);
 	g_print ("description: %s\n", description);
-	ch_util_show_calibration (calibration);
+	ch_util_show_calibration (&calibration);
 out:
 	return ret;
 }
@@ -542,11 +633,12 @@ out:
 static gboolean
 ch_util_set_calibration (ChUtilPrivate *priv, gchar **values, GError **error)
 {
+	CdMat3x3 calibration;
 	gboolean ret;
-	gdouble calibration[9];
+	gdouble *calibration_tmp;
 	guint16 calibration_index = 0;
-	guint types = 0;
 	guint i;
+	guint types = 0;
 
 	/* parse */
 	if (g_strv_length (values) != 12) {
@@ -555,7 +647,7 @@ ch_util_set_calibration (ChUtilPrivate *priv, gchar **values, GError **error)
 				     "invalid input, expect 'index' 'types' 'values' 'description'");
 		goto out;
 	}
-	calibration_index = atoi (values[0]);
+	calibration_index = g_ascii_strtoull (values[0], NULL, 10);
 
 	/* try to parse magic constants */
 	if (g_strstr_len (values[1], -1, "lcd") != NULL)
@@ -570,12 +662,13 @@ ch_util_set_calibration (ChUtilPrivate *priv, gchar **values, GError **error)
 				     "invalid type, expected 'lcd', 'crt', 'projector'");
 		goto out;
 	}
+	calibration_tmp = cd_mat33_get_data (&calibration);
 	for (i = 0; i < 9; i++)
-		calibration[i] = atof (values[i+2]);
+		calibration_tmp[i] = g_ascii_strtod (values[i+2], NULL);
 
 	/* check is valid */
 	for (i = 0; i < 9; i++) {
-		if (calibration[i] > 0x7fff || calibration[i] < -0x7fff) {
+		if (calibration_tmp[i] > 0x7fff || calibration_tmp[i] < -0x7fff) {
 			ret = FALSE;
 			g_set_error_literal (error, 1, 0,
 					     "invalid value, expect -1.0 to +1.0");
@@ -595,13 +688,13 @@ ch_util_set_calibration (ChUtilPrivate *priv, gchar **values, GError **error)
 	/* set to HW */
 	ret = ch_device_cmd_set_calibration (priv->device,
 					     calibration_index,
-					     calibration,
+					     &calibration,
 					     types,
 					     values[11],
 					     error);
 	if (!ret)
 		goto out;
-	ch_util_show_calibration (calibration);
+	ch_util_show_calibration (&calibration);
 out:
 	return ret;
 }
@@ -622,7 +715,7 @@ ch_util_clear_calibration (ChUtilPrivate *priv, gchar **values, GError **error)
 				     "invalid input, expect 'index'");
 		goto out;
 	}
-	calibration_index = atoi (values[0]);
+	calibration_index = g_ascii_strtoull (values[0], NULL, 10);
 
 	/* set to HW */
 	ret = ch_device_cmd_clear_calibration (priv->device,
@@ -693,7 +786,7 @@ ch_util_set_calibration_ccmx (ChUtilPrivate *priv, gchar **values, GError **erro
 	}
 
 	/* load file */
-	calibration_index = atoi (values[0]);
+	calibration_index = g_ascii_strtoull (values[0], NULL, 10);
 
 	/* set to HW */
 	ret = ch_device_cmd_set_calibration_ccmx (priv->device,
@@ -819,7 +912,7 @@ ch_util_set_serial_number (ChUtilPrivate *priv, gchar **values, GError **error)
 			goto out;
 		}
 	} else {
-		serial_number = atol (values[0]);
+		serial_number = g_ascii_strtoull (values[0], NULL, 10);
 	}
 	if (serial_number == 0) {
 		ret = FALSE;
@@ -838,6 +931,104 @@ out:
 	return ret;
 }
 
+/**
+ * ch_util_get_owner_name:
+ **/
+static gboolean
+ch_util_get_owner_name (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gchar name[CH_OWNER_LENGTH_MAX];
+
+	/* get from HW */
+	ret = ch_device_cmd_get_owner_name (priv->device, name, error);
+	if (!ret)
+		goto out;
+	g_print ("%s\n", name);
+out:
+	return ret;
+}
+
+/**
+ * ch_util_set_owner_name:
+ **/
+static gboolean
+ch_util_set_owner_name (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gchar name[CH_OWNER_LENGTH_MAX];
+
+	/* parse */
+	if (g_strv_length (values) != 1) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect 'name'");
+		goto out;
+	}
+	if(strlen(values[0]) >= CH_OWNER_LENGTH_MAX) {
+		g_print ("truncating name to %d characters\n", CH_OWNER_LENGTH_MAX-1);
+	}
+	memset(name, 0, CH_OWNER_LENGTH_MAX);
+	g_strlcpy(name, values[0], CH_OWNER_LENGTH_MAX);
+
+	/* set to HW */
+	g_print ("setting name to %s\n", name);
+	ret = ch_device_cmd_set_owner_name (priv->device, name, error);
+	if (!ret)
+		goto out;
+out:
+	return ret;
+}
+
+/**
+ * ch_util_get_owner_email:
+ **/
+static gboolean
+ch_util_get_owner_email (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gchar email[CH_OWNER_LENGTH_MAX];
+
+	/* get from HW */
+	ret = ch_device_cmd_get_owner_email (priv->device, email, error);
+	if (!ret)
+		goto out;
+	g_print ("%s\n", email);
+out:
+	return ret;
+}
+
+/**
+ * ch_util_set_owner_email:
+ **/
+static gboolean
+ch_util_set_owner_email (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gchar email[CH_OWNER_LENGTH_MAX];
+
+	/* parse */
+	if (g_strv_length (values) != 1) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect 'email'");
+		goto out;
+	}
+	if(strlen(values[0]) >= CH_OWNER_LENGTH_MAX) {
+		g_print ("truncating email to %d characters\n", CH_OWNER_LENGTH_MAX-1);
+	}
+	memset (email, 0, CH_OWNER_LENGTH_MAX);
+	g_strlcpy (email, values[0], CH_OWNER_LENGTH_MAX);
+
+	/* set to HW */
+	g_print ("setting email to %s\n", email);
+	ret = ch_device_cmd_set_owner_email (priv->device, email, error);
+	if (!ret)
+		goto out;
+out:
+	return ret;
+}
+
 /**
  * ch_util_get_leds:
  **/
@@ -845,7 +1036,7 @@ static gboolean
 ch_util_get_leds (ChUtilPrivate *priv, gchar **values, GError **error)
 {
 	gboolean ret;
-	ChStatusLed leds = 0xff;
+	guint8 leds = 0;
 
 	/* get from HW */
 	ret = ch_device_cmd_get_leds (priv->device, &leds, error);
@@ -870,7 +1061,7 @@ static gboolean
 ch_util_set_leds (ChUtilPrivate *priv, gchar **values, GError **error)
 {
 	gboolean ret;
-	ChStatusLed leds;
+	guint8 leds;
 	guint8 repeat = 0;
 	guint8 time_on = 0x00;
 	guint8 time_off = 0x00;
@@ -894,7 +1085,7 @@ ch_util_set_leds (ChUtilPrivate *priv, gchar **values, GError **error)
 	} else if (g_strcmp0 (values[0], "both") == 0) {
 		leds = CH_STATUS_LED_RED | CH_STATUS_LED_GREEN;
 	} else {
-		leds = atoi (values[0]);
+		leds = g_ascii_strtoull (values[0], NULL, 10);
 		if (leds > 3) {
 			ret = FALSE;
 			g_set_error (error, 1, 0,
@@ -906,9 +1097,9 @@ ch_util_set_leds (ChUtilPrivate *priv, gchar **values, GError **error)
 
 	/* get the optional other parameters */
 	if (g_strv_length (values) == 4) {
-		repeat = atoi (values[1]);
-		time_on = atoi (values[2]);
-		time_off = atoi (values[3]);
+		repeat = g_ascii_strtoull (values[1], NULL, 10);
+		time_on = g_ascii_strtoull (values[2], NULL, 10);
+		time_off = g_ascii_strtoull (values[3], NULL, 10);
 	}
 
 	/* set to HW */
@@ -924,6 +1115,76 @@ out:
 	return ret;
 }
 
+/**
+ * ch_util_get_pcb_errata:
+ **/
+static gboolean
+ch_util_get_pcb_errata (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	guint16 pcb_errata = CH_PCB_ERRATA_NONE;
+
+	/* get from HW */
+	ret = ch_device_cmd_get_pcb_errata (priv->device, &pcb_errata, error);
+	if (!ret)
+		goto out;
+	if (pcb_errata == 0) {
+		g_print ("Errata: none\n");
+		goto out;
+	}
+	if ((pcb_errata & CH_PCB_ERRATA_SWAPPED_LEDS) > 0)
+		g_print ("Errata: swapped-leds\n");
+out:
+	return ret;
+}
+
+/**
+ * ch_util_set_pcb_errata:
+ **/
+static gboolean
+ch_util_set_pcb_errata (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	guint16 pcb_errata = CH_PCB_ERRATA_NONE;
+
+	/* parse */
+	if (g_strv_length (values) != 1) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect 'none|swapped-leds'");
+		goto out;
+	}
+
+	/* swap the green and red LEDs */
+	if (g_strstr_len (values[0], -1, "swapped-leds") != NULL) {
+		g_print ("Errata: swapped-leds\n");
+		pcb_errata += CH_PCB_ERRATA_SWAPPED_LEDS;
+	}
+
+	/* nothing known by this client version */
+	if (pcb_errata == CH_PCB_ERRATA_NONE)
+		pcb_errata = g_ascii_strtoull (values[0], NULL, 10);
+
+	if (pcb_errata == 0)
+		g_print ("Errata: none\n");
+
+	/* set to HW */
+	ret = ch_device_cmd_set_pcb_errata (priv->device,
+					    pcb_errata,
+					    error);
+	if (!ret)
+		goto out;
+
+	/* save EEPROM */
+	ret = ch_device_cmd_write_eeprom (priv->device,
+					  CH_WRITE_EEPROM_MAGIC,
+					  error);
+	if (!ret)
+		goto out;
+out:
+	return ret;
+}
+
 /**
  * ch_util_get_dark_offsets:
  **/
@@ -939,7 +1200,85 @@ ch_util_get_dark_offsets (ChUtilPrivate *priv, gchar **values, GError **error)
 					      error);
 	if (!ret)
 		goto out;
-	g_print ("R:%.4f G:%.4f B:%.4f\n", value.R, value.G, value.B);
+	g_print ("R:%.5f G:%.5f B:%.5f\n", value.R, value.G, value.B);
+out:
+	return ret;
+}
+
+/**
+ * ch_util_set_dark_offsets_auto:
+ **/
+static gboolean
+ch_util_set_dark_offsets_auto (ChUtilPrivate *priv, GError **error)
+{
+	gboolean ret;
+	CdColorRGB value_old;
+	CdColorRGB value;
+
+	/* get from HW */
+	ret = ch_device_cmd_get_dark_offsets (priv->device,
+					      &value_old,
+					      error);
+	if (!ret)
+		goto out;
+
+	/* set dark offsets */
+	cd_color_set_rgb (&value, 0.0f, 0.0f, 0.0f);
+
+	/* set to HW */
+	ret = ch_device_cmd_set_dark_offsets (priv->device,
+					      &value,
+					      error);
+	if (!ret)
+		goto out;
+
+	/* setup HW */
+	ret = ch_device_cmd_set_integral_time (priv->device,
+					       CH_INTEGRAL_TIME_VALUE_MAX,
+					       error);
+	if (!ret)
+		goto out;
+	ret = ch_device_cmd_set_multiplier (priv->device,
+					    CH_FREQ_SCALE_100,
+					    error);
+	if (!ret)
+		goto out;
+
+	/* get from HW */
+	ret = ch_device_cmd_take_readings (priv->device,
+					   &value,
+					   error);
+	if (!ret)
+		goto out;
+	g_print ("Values: R:%.5f G:%.5f B:%.5f\n", value.R, value.G, value.B);
+
+	/* TRANSLATORS: ask before we set these */
+	ret = ch_util_get_prompt (_("Set these values as the dark offsets"), FALSE);
+	if (!ret) {
+		/* restore HW */
+		ret = ch_device_cmd_set_dark_offsets (priv->device,
+						      &value_old,
+						      error);
+		if (!ret)
+			goto out;
+		g_set_error_literal (error, 1, 0,
+				     "user declined");
+		goto out;
+	}
+
+	/* set to HW */
+	ret = ch_device_cmd_set_dark_offsets (priv->device,
+					      &value,
+					      error);
+	if (!ret)
+		goto out;
+
+	/* save EEPROM */
+	ret = ch_device_cmd_write_eeprom (priv->device,
+					  CH_WRITE_EEPROM_MAGIC,
+					  error);
+	if (!ret)
+		goto out;
 out:
 	return ret;
 }
@@ -953,6 +1292,12 @@ ch_util_set_dark_offsets (ChUtilPrivate *priv, gchar **values, GError **error)
 	gboolean ret;
 	CdColorRGB value;
 
+	/* be interactive */
+	if (g_strv_length (values) == 0) {
+		ret = ch_util_set_dark_offsets_auto (priv, error);
+		goto out;
+	}
+
 	/* parse */
 	if (g_strv_length (values) != 3) {
 		ret = FALSE;
@@ -960,9 +1305,9 @@ ch_util_set_dark_offsets (ChUtilPrivate *priv, gchar **values, GError **error)
 				     "invalid input, expect 'value'");
 		goto out;
 	}
-	value.R = atof (values[0]);
-	value.G = atof (values[1]);
-	value.B = atof (values[2]);
+	value.R = g_ascii_strtod (values[0], NULL);
+	value.G = g_ascii_strtod (values[1], NULL);
+	value.B = g_ascii_strtod (values[2], NULL);
 
 	/* set to HW */
 	ret = ch_device_cmd_set_dark_offsets (priv->device,
@@ -1004,14 +1349,45 @@ out:
 static gboolean
 ch_util_take_reading_raw (ChUtilPrivate *priv, gchar **values, GError **error)
 {
+	ChColorSelect color_select = 0;
+	ChFreqScale multiplier = 0;
 	gboolean ret;
+	guint16 integral_time = 0;
 	guint16 take_reading;
 
+	/* get from HW */
+	ret = ch_device_cmd_get_color_select (priv->device, &color_select, error);
+	if (!ret)
+		goto out;
+
+	/* TRANSLATORS: this is the enabled sensor color */
+	g_print ("%s:\t\t%s\n", _("Color"),
+		 ch_color_select_to_string (color_select));
+
+	/* get from HW */
+	ret = ch_device_cmd_get_multiplier (priv->device, &multiplier, error);
+	if (!ret)
+		goto out;
+
+	/* TRANSLATORS: this is the sensor scale factor */
+	g_print ("%s:\t%s\n", _("Multiplier"),
+		 ch_multiplier_to_string (multiplier));
+
+	/* get from HW */
+	ret = ch_device_cmd_get_integral_time (priv->device, &integral_time, error);
+	if (!ret)
+		goto out;
+
+	/* TRANSLATORS: this is the sensor sample time */
+	g_print ("%s:\t0x%04x\n", _("Integral"), integral_time);
+
 	/* get from HW */
 	ret = ch_device_cmd_take_reading_raw (priv->device, &take_reading, error);
 	if (!ret)
 		goto out;
-	g_print ("%i\n", take_reading);
+
+	/* TRANSLATORS: this is the number of pulses detected */
+	g_print ("%s:\t\t%i\n", _("Pulses"), take_reading);
 out:
 	return ret;
 }
@@ -1022,8 +1398,27 @@ out:
 static gboolean
 ch_util_take_readings (ChUtilPrivate *priv, gchar **values, GError **error)
 {
-	gboolean ret;
 	CdColorRGB value;
+	ChFreqScale multiplier = 0;
+	gboolean ret;
+	guint16 integral_time = 0;
+
+	/* get from HW */
+	ret = ch_device_cmd_get_multiplier (priv->device, &multiplier, error);
+	if (!ret)
+		goto out;
+
+	/* TRANSLATORS: this is the sensor scale factor */
+	g_print ("%s:\t%s\n", _("Multiplier"),
+		 ch_multiplier_to_string (multiplier));
+
+	/* get from HW */
+	ret = ch_device_cmd_get_integral_time (priv->device, &integral_time, error);
+	if (!ret)
+		goto out;
+
+	/* TRANSLATORS: this is the sensor sample time */
+	g_print ("%s:\t0x%04x\n", _("Integral"), integral_time);
 
 	/* get from HW */
 	ret = ch_device_cmd_take_readings (priv->device,
@@ -1031,7 +1426,7 @@ ch_util_take_readings (ChUtilPrivate *priv, gchar **values, GError **error)
 					   error);
 	if (!ret)
 		goto out;
-	g_print ("R:%.4f G:%.4f B:%.4f\n", value.R, value.G, value.B);
+	g_print ("R:%.5f G:%.5f B:%.5f\n", value.R, value.G, value.B);
 out:
 	return ret;
 }
@@ -1042,9 +1437,11 @@ out:
 static gboolean
 ch_util_take_readings_xyz (ChUtilPrivate *priv, gchar **values, GError **error)
 {
-	gboolean ret;
 	CdColorXYZ value;
+	ChFreqScale multiplier = 0;
+	gboolean ret;
 	guint16 calibration_index = 0;
+	guint16 integral_time = 0;
 
 	/* parse */
 	if (g_strv_length (values) != 1) {
@@ -1053,7 +1450,24 @@ ch_util_take_readings_xyz (ChUtilPrivate *priv, gchar **values, GError **error)
 				     "invalid input, expect 'calibration_index'");
 		goto out;
 	}
-	calibration_index = atoi (values[0]);
+	calibration_index = g_ascii_strtoull (values[0], NULL, 10);
+
+	/* get from HW */
+	ret = ch_device_cmd_get_multiplier (priv->device, &multiplier, error);
+	if (!ret)
+		goto out;
+
+	/* TRANSLATORS: this is the sensor scale factor */
+	g_print ("%s:\t%s\n", _("Multiplier"),
+		 ch_multiplier_to_string (multiplier));
+
+	/* get from HW */
+	ret = ch_device_cmd_get_integral_time (priv->device, &integral_time, error);
+	if (!ret)
+		goto out;
+
+	/* TRANSLATORS: this is the sensor sample time */
+	g_print ("%s:\t0x%04x\n", _("Integral"), integral_time);
 
 	/* get from HW */
 	ret = ch_device_cmd_take_readings_xyz (priv->device,
@@ -1062,7 +1476,7 @@ ch_util_take_readings_xyz (ChUtilPrivate *priv, gchar **values, GError **error)
 					       error);
 	if (!ret)
 		goto out;
-	g_print ("X:%.4f Y:%.4f Z:%.4f\n", value.X, value.Y, value.Z);
+	g_print ("X:%.5f Y:%.5f Z:%.5f\n", value.X, value.Y, value.Z);
 out:
 	return ret;
 }
@@ -1085,10 +1499,10 @@ out:
 }
 
 /**
- * ch_util_flash_firmware:
+ * ch_util_flash_firmware_force:
  **/
 static gboolean
-ch_util_flash_firmware (ChUtilPrivate *priv, gchar **values, GError **error)
+ch_util_flash_firmware_force (ChUtilPrivate *priv, gchar **values, GError **error)
 {
 	gboolean ret;
 
@@ -1106,6 +1520,49 @@ ch_util_flash_firmware (ChUtilPrivate *priv, gchar **values, GError **error)
 					error);
 	if (!ret)
 		goto out;
+
+	/* print success */
+	g_print ("INFO: Flashing was successful.\n");
+out:
+	return ret;
+}
+
+/**
+ * ch_util_flash_firmware:
+ **/
+static gboolean
+ch_util_flash_firmware (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+
+	/* parse */
+	if (g_strv_length (values) != 1) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect 'filename'");
+		goto out;
+	}
+
+	/* print warning */
+	g_print ("WARNING: Do not shutdown the computer or unplug the device.\n");
+
+	/* TRANSLATORS: confirmation */
+	ret = ch_util_get_prompt (_("Flash the device?"), FALSE);
+	if (!ret) {
+		g_set_error_literal (error, 1, 0,
+				     "user declined");
+		goto out;
+	}
+
+	/* set to HW */
+	ret = ch_client_flash_firmware (priv->client,
+					values[0],
+					error);
+	if (!ret)
+		goto out;
+
+	/* print success */
+	g_print ("INFO: Flashing was successful.\n");
 out:
 	return ret;
 }
@@ -1146,7 +1603,7 @@ ch_util_set_pre_scale (ChUtilPrivate *priv, gchar **values, GError **error)
 				     "invalid input, expect 'value'");
 		goto out;
 	}
-	pre_scale = atof (values[0]);
+	pre_scale = g_ascii_strtod (values[0], NULL);
 	if (pre_scale < -0x7fff || pre_scale > 0x7fff) {
 		ret = FALSE;
 		g_set_error (error, 1, 0,
@@ -1201,7 +1658,7 @@ ch_util_set_post_scale (ChUtilPrivate *priv, gchar **values, GError **error)
 				     "invalid input, expect 'value'");
 		goto out;
 	}
-	post_scale = atof (values[0]);
+	post_scale = g_ascii_strtod (values[0], NULL);
 	if (post_scale < -0x7fff || post_scale > 0x7fff) {
 		ret = FALSE;
 		g_set_error (error, 1, 0,
@@ -1253,7 +1710,7 @@ ch_util_set_flash_success (ChUtilPrivate *priv, gchar **values, GError **error)
 				     "invalid input, expect 'value'");
 		goto out;
 	}
-	flash_success = atoi (values[0]);
+	flash_success = g_ascii_strtoull (values[0], NULL, 10);
 	if (flash_success > 1) {
 		ret = FALSE;
 		g_set_error (error, 1, 0,
@@ -1272,6 +1729,170 @@ out:
 	return ret;
 }
 
+/**
+ * ch_util_eeprom_write:
+ **/
+static gboolean
+ch_util_eeprom_write (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gsize len;
+	guint16 address;
+	guint8 *data = NULL;
+	guint i;
+
+	/* parse */
+	if (g_strv_length (values) != 2) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect 'address (base-16)' 'length (base-10)'");
+		goto out;
+	}
+
+	/* read flash */
+	address = g_ascii_strtoull (values[0], NULL, 16);
+	if (address < CH_EEPROM_ADDR_RUNCODE) {
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "invalid address 0x%04x",
+			     address);
+		goto out;
+	}
+	len = g_ascii_strtoull (values[1], NULL, 10);
+	if (len < 1 || len > 60) {
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "invalid length %" G_GSIZE_FORMAT " (1-60)",
+			     len);
+		goto out;
+	}
+
+	/* just write zeros */
+	data = g_new0 (guint8, len);
+	ret = ch_device_cmd_write_flash (priv->device,
+					 address,
+					 data,
+					 len,
+					 error);
+	if (!ret)
+		goto out;
+
+	/* flush */
+	ret = ch_device_cmd_write_flash (priv->device,
+					 address | CH_FLASH_TRANSFER_BLOCK_SIZE,
+					 data,
+					 len,
+					 error);
+	if (!ret)
+		goto out;
+
+	g_print ("Wrote:\n");
+	for (i=0; i< len; i++)
+		g_print ("0x%04x = %02x\n", address + i, data[i]);
+out:
+	g_free (data);
+	return ret;
+}
+
+/**
+ * ch_util_eeprom_erase:
+ **/
+static gboolean
+ch_util_eeprom_erase (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gsize len;
+	guint16 address;
+
+	/* parse */
+	if (g_strv_length (values) != 2) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect 'address (base-16)' 'length (base-10)'");
+		goto out;
+	}
+
+	/* read flash */
+	address = g_ascii_strtoull (values[0], NULL, 16);
+	if (address < CH_EEPROM_ADDR_RUNCODE) {
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "invalid address 0x%04x",
+			     address);
+		goto out;
+	}
+	len = g_ascii_strtoull (values[1], NULL, 10);
+	if (len < 1 || len > 0xffff) {
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "invalid length %" G_GSIZE_FORMAT " (1-60)",
+			     len);
+		goto out;
+	}
+	ret = ch_device_cmd_erase_flash (priv->device,
+					 address,
+					 len,
+					 error);
+	if (!ret)
+		goto out;
+out:
+	return ret;
+}
+
+/**
+ * ch_util_eeprom_read:
+ **/
+static gboolean
+ch_util_eeprom_read (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gsize len;
+	guint16 address;
+	guint8 *data = NULL;
+	guint i;
+
+	/* parse */
+	if (g_strv_length (values) != 2) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect 'address (base-16)' 'length (base-10)'");
+		goto out;
+	}
+
+	/* read flash */
+	address = g_ascii_strtoull (values[0], NULL, 16);
+	if (address < CH_EEPROM_ADDR_RUNCODE) {
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "invalid address 0x%04x",
+			     address);
+		goto out;
+	}
+	len = g_ascii_strtoull (values[1], NULL, 10);
+	if (len < 1 || len > 60) {
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "invalid length %" G_GSIZE_FORMAT " (1-60)",
+			     len);
+		goto out;
+	}
+	data = g_new0 (guint8, len);
+	ret = ch_device_cmd_read_flash (priv->device,
+					address,
+					data,
+					len,
+					error);
+	if (!ret)
+		goto out;
+
+	g_print ("Read:\n");
+	for (i=0; i< len; i++)
+		g_print ("0x%04x = %02x\n", address + i, data[i]);
+out:
+	g_free (data);
+	return ret;
+}
+
 /**
  * ch_util_ignore_cb:
  **/
@@ -1395,6 +2016,26 @@ main (int argc, char *argv[])
 		     /* TRANSLATORS: command description */
 		     _("Sets the sensor serial number"),
 		     ch_util_set_serial_number);
+	ch_util_add (priv->cmd_array,
+		     "get-owner-name",
+		     /* TRANSLATORS: command description */
+		     _("Gets the owner's name"),
+		     ch_util_get_owner_name);
+	ch_util_add (priv->cmd_array,
+		     "set-owner-name",
+		     /* TRANSLATORS: command description */
+		     _("Sets the owner's name"),
+		     ch_util_set_owner_name);
+	ch_util_add (priv->cmd_array,
+		     "get-owner-email",
+		     /* TRANSLATORS: command description */
+		     _("Gets the owner's email address"),
+		     ch_util_get_owner_email);
+	ch_util_add (priv->cmd_array,
+		     "set-owner-email",
+		     /* TRANSLATORS: command description */
+		     _("Sets the owner's email address"),
+		     ch_util_set_owner_email);
 	ch_util_add (priv->cmd_array,
 		     "get-leds",
 		     /* TRANSLATORS: command description */
@@ -1405,6 +2046,16 @@ main (int argc, char *argv[])
 		     /* TRANSLATORS: command description */
 		     _("Sets the LEDs"),
 		     ch_util_set_leds);
+	ch_util_add (priv->cmd_array,
+		     "get-pcb-errata",
+		     /* TRANSLATORS: command description */
+		     _("Gets the PCB errata"),
+		     ch_util_get_pcb_errata);
+	ch_util_add (priv->cmd_array,
+		     "set-pcb-errata",
+		     /* TRANSLATORS: command description */
+		     _("Sets the PCB errata"),
+		     ch_util_set_pcb_errata);
 	ch_util_add (priv->cmd_array,
 		     "get-dark-offsets",
 		     /* TRANSLATORS: command description */
@@ -1445,6 +2096,26 @@ main (int argc, char *argv[])
 		     /* TRANSLATORS: command description */
 		     _("Flash firmware into the processor"),
 		     ch_util_flash_firmware);
+	ch_util_add (priv->cmd_array,
+		     "flash-firmware-force",
+		     /* TRANSLATORS: command description */
+		     _("Flash firmware into the processor"),
+		     ch_util_flash_firmware_force);
+	ch_util_add (priv->cmd_array,
+		     "eeprom-read",
+		     /* TRANSLATORS: command description */
+		     _("Read EEPROM at a specified address"),
+		     ch_util_eeprom_read);
+	ch_util_add (priv->cmd_array,
+		     "eeprom-erase",
+		     /* TRANSLATORS: command description */
+		     _("Erase EEPROM at a specified address"),
+		     ch_util_eeprom_erase);
+	ch_util_add (priv->cmd_array,
+		     "eeprom-write",
+		     /* TRANSLATORS: command description */
+		     _("Write EEPROM at a specified address"),
+		     ch_util_eeprom_write);
 	ch_util_add (priv->cmd_array,
 		     "get-pre-scale",
 		     /* TRANSLATORS: command description */
@@ -1490,6 +2161,11 @@ main (int argc, char *argv[])
 		     /* TRANSLATORS: command description */
 		     _("Gets the hardware version"),
 		     ch_util_get_hardware_version);
+	ch_util_add (priv->cmd_array,
+		     "take-reading-array",
+		     /* TRANSLATORS: command description */
+		     _("Gets an array of raw samples"),
+		     ch_util_take_reading_array);
 
 	/* sort by command name */
 	g_ptr_array_sort (priv->cmd_array,
