@@ -41,6 +41,7 @@ ch_test_math_convert_func (void)
 	/* test converting to packed struct */
 	value = 3.1415927f;
 	ch_double_to_packed_float (value, &pf);
+	pf.raw = ch_packed_float_get_value (&pf);
 	g_assert_cmpint (pf.offset, ==, 3);
 	g_assert_cmpint (pf.fraction, <, 0x249f);
 	g_assert_cmpint (pf.fraction, >, 0x240f);
@@ -48,6 +49,7 @@ ch_test_math_convert_func (void)
 	/* test converting to packed struct */
 	value = -3.1415927f;
 	ch_double_to_packed_float (value, &pf);
+	pf.raw = ch_packed_float_get_value (&pf);
 	g_assert_cmpint (pf.offset, ==, -4);
 	g_assert_cmpint (pf.fraction, <, (0x240f ^ 0xffff));
 	g_assert_cmpint (pf.fraction, >, (0x249f ^ 0xffff));
@@ -55,6 +57,7 @@ ch_test_math_convert_func (void)
 	/* test converting positive to float */
 	pf.offset = 3;
 	pf.fraction = 0x243c;
+	pf.raw = ch_packed_float_get_value (&pf);
 	ch_packed_float_to_double (&pf, &value);
 	g_assert_cmpfloat (value, >, 3.1415);
 	g_assert_cmpfloat (value, <, 3.1416);
@@ -63,14 +66,17 @@ ch_test_math_convert_func (void)
 	pf.offset = -4;
 	pf.fraction = 0x243b ^ 0xffff;
 	ch_packed_float_to_double (&pf, &value);
+	pf.raw = ch_packed_float_get_value (&pf);
 	g_assert_cmpfloat (value, >, -3.1416);
 	g_assert_cmpfloat (value, <, -3.1415);
 
 	/* test converting zero */
 	value = 0.0f;
 	ch_double_to_packed_float (value, &pf);
+	pf.raw = ch_packed_float_get_value (&pf);
 	g_assert_cmpint (pf.offset, ==, 0);
 	g_assert_cmpint (pf.fraction, ==, 0);
+	ch_packed_float_set_value (&pf, pf.raw);
 	ch_packed_float_to_double (&pf, &value);
 	g_assert_cmpfloat (value, >, -0.001f);
 	g_assert_cmpfloat (value, <, +0.001f);
@@ -78,9 +84,11 @@ ch_test_math_convert_func (void)
 	/* test converting positive */
 	value = +1.4f;
 	ch_double_to_packed_float (value, &pf);
+	pf.raw = ch_packed_float_get_value (&pf);
 	g_assert_cmpint (pf.offset, ==, 1);
 	g_assert_cmpint (pf.fraction, <, 0x6668);
 	g_assert_cmpint (pf.fraction, >, 0x6663);
+	ch_packed_float_set_value (&pf, pf.raw);
 	ch_packed_float_to_double (&pf, &value);
 	g_assert_cmpfloat (value, <, 1.41);
 	g_assert_cmpfloat (value, >, 1.39);
@@ -88,9 +96,11 @@ ch_test_math_convert_func (void)
 	/* test converting negative */
 	value = -1.4f;
 	ch_double_to_packed_float (value, &pf);
+	pf.raw = ch_packed_float_get_value (&pf);
 	g_assert_cmpint (pf.offset, ==, -2);
 	g_assert_cmpint (pf.fraction, <, (0x6662 ^ 0xffff));
 	g_assert_cmpint (pf.fraction, >, (0x6668 ^ 0xffff));
+	ch_packed_float_set_value (&pf, pf.raw);
 	ch_packed_float_to_double (&pf, &value);
 	g_assert_cmpfloat (value, <, -1.39);
 	g_assert_cmpfloat (value, >, -1.41);
@@ -98,8 +108,10 @@ ch_test_math_convert_func (void)
 	/* test converting negative max */
 	value = -0x7fff;
 	ch_double_to_packed_float (value, &pf);
+	pf.raw = ch_packed_float_get_value (&pf);
 	g_assert_cmpint (pf.offset, ==, -32767);
 	g_assert_cmpint (pf.fraction, ==, 0);
+	ch_packed_float_set_value (&pf, pf.raw);
 	ch_packed_float_to_double (&pf, &value);
 	g_assert_cmpfloat (value, >, -32768.0001);
 	g_assert_cmpfloat (value, <, +32767.9999);
@@ -252,7 +264,7 @@ ch_test_state_func (void)
 	gboolean ret;
 	GError *error = NULL;
 	guint16 integral_time = 0;
-	ChStatusLed leds = 0;
+	guint8 leds = 0;
 	GUsbDevice *device;
 
 	/* new device */
@@ -344,8 +356,8 @@ ch_test_eeprom_func (void)
 	gdouble pre_scale = 0;
 	gdouble pre_scale_tmp = 0;
 	guint32 serial_number = 0;
-	gdouble calibration[9];
-	gdouble calibration_tmp[9];
+	CdMat3x3 calibration;
+	CdMat3x3 calibration_tmp;
 	gchar desc[24];
 	GUsbDevice *device;
 
@@ -427,18 +439,18 @@ ch_test_eeprom_func (void)
 	g_assert_cmpint (value.B, ==, 0.56);
 
 	/* verify calibration */
-	calibration[0] = 1.0f;
-	calibration[1] = 2.0f;
-	calibration[2] = 3.0f;
-	calibration[3] = 4.0f;
-	calibration[4] = 5.0f;
-	calibration[5] = 6.0f;
-	calibration[6] = 7.0f;
-	calibration[7] = 8.0f;
-	calibration[8] = 9.0f;
+	calibration.m00 = 1.0f;
+	calibration.m01 = 2.0f;
+	calibration.m02 = 3.0f;
+	calibration.m10 = 4.0f;
+	calibration.m11 = 5.0f;
+	calibration.m12 = 6.0f;
+	calibration.m20 = 7.0f;
+	calibration.m21 = 8.0f;
+	calibration.m22 = 9.0f;
 	ret = ch_device_cmd_set_calibration (device,
 					     60,
-					     calibration,
+					     &calibration,
 					     CH_CALIBRATION_TYPE_CRT,
 					     "test0",
 					     &error);
@@ -446,7 +458,7 @@ ch_test_eeprom_func (void)
 	g_assert (ret);
 	ret = ch_device_cmd_set_calibration (device,
 					     61,
-					     calibration,
+					     &calibration,
 					     CH_CALIBRATION_TYPE_PROJECTOR,
 					     "test1",
 					     &error);
@@ -454,7 +466,7 @@ ch_test_eeprom_func (void)
 	g_assert (ret);
 	ret = ch_device_cmd_set_calibration (device,
 					     60,
-					     calibration,
+					     &calibration,
 					     CH_CALIBRATION_TYPE_CRT,
 					     "test0",
 					     &error);
@@ -464,27 +476,27 @@ ch_test_eeprom_func (void)
 	/* read back data */
 	ret = ch_device_cmd_get_calibration (device,
 					     60,
-					     calibration_tmp,
+					     &calibration_tmp,
 					     &types,
 					     desc,
 					     &error);
 	g_assert_no_error (error);
 	g_assert (ret);
-	g_assert (memcmp (calibration_tmp,
-			  calibration,
+	g_assert (memcmp (&calibration_tmp,
+			  &calibration,
 			  sizeof (gfloat) * 9) == 0);
 	g_assert_cmpint (types, ==, CH_CALIBRATION_TYPE_CRT);
 	g_assert_cmpstr (desc, ==, "test0");
 	ret = ch_device_cmd_get_calibration (device,
 					     61,
-					     calibration_tmp,
+					     &calibration_tmp,
 					     &types,
 					     desc,
 					     &error);
 	g_assert_no_error (error);
 	g_assert (ret);
-	g_assert (memcmp (calibration_tmp,
-			  calibration,
+	g_assert (memcmp (&calibration_tmp,
+			  &calibration,
 			  sizeof (gfloat) * 9) == 0);
 	g_assert_cmpint (types, ==, CH_CALIBRATION_TYPE_PROJECTOR);
 	g_assert_cmpstr (desc, ==, "test1");
@@ -594,7 +606,7 @@ ch_test_reading_xyz_func (void)
 {
 	ChClient *client;
 	gboolean ret;
-	gdouble calibration[9];
+	CdMat3x3 calibration;
 	CdColorXYZ reading1;
 	CdColorXYZ reading2;
 	gdouble scaling_factor_actual;
@@ -621,18 +633,10 @@ ch_test_reading_xyz_func (void)
 	g_assert (device != NULL);
 
 	/* set unity calibration */
-	calibration[0] = 1.0f;
-	calibration[1] = 0.0f;
-	calibration[2] = 0.0f;
-	calibration[3] = 0.0f;
-	calibration[4] = 1.0f;
-	calibration[5] = 0.0f;
-	calibration[6] = 0.0f;
-	calibration[7] = 0.0f;
-	calibration[8] = 1.0f;
+	cd_mat33_set_identity (&calibration);
 	ret = ch_device_cmd_set_calibration (device,
 					     60,
-					     calibration,
+					     &calibration,
 					     CH_CALIBRATION_TYPE_ALL,
 					     "test0",
 					     &error);
@@ -735,7 +739,6 @@ int
 main (int argc, char **argv)
 {
 	g_type_init ();
-	g_thread_init (NULL);
 	g_test_init (&argc, &argv, NULL);
 
 	/* only critical and error are fatal */
